#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta
from itertools import chain
import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    nearest_workday,
    next_monday,
    previous_friday,
    sunday_to_monday,
)
from pytz import timezone

from .common_holidays import (
    european_labour_day,
    new_years_day,
)
from .lunisolar_holidays import (
    chinese_lunar_new_year_dates,
    qingming_festival_dates,
    dragon_boat_festival_dates,
    mid_autumn_festival_dates,
)
from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
    TUESDAY,
    THURSDAY,
    SATURDAY,
    SUNDAY,
)


def before_chinese_new_year_offset(holidays):
    """
    For Holidays that come before Chinese New Year, we subtract a day
    and then move any weekends to previous friday.
    """
    return pd.to_datetime(
        holidays.map(
            lambda d: previous_friday(d)
        )
    )


def chinese_new_year_offset(holidays):
    """
    For Holidays on or after Chinese New Year, we add a day
    and then move any weekends to next monday.
    """
    return pd.to_datetime(
        holidays.map(
            lambda d: next_monday(d)
        )
    )


def nearest_workday_after_2013(dt):
    """
    Nearest workday starting in 2014.
    """
    return nearest_workday(dt) if dt.year > 2013 else dt


def manual_nearest_workday(holidays):
    """
    Nearest workday observance rule for Chinese lunar calendar holidays.
    The nearest workday rule seems to start in 2014 for these holidays.
    """
    return pd.to_datetime(
        holidays.map(
            lambda d: nearest_workday_after_2013(d)
        )
    )


def manual_extra_days(holidays):
    """
    Four day weekend makeup days for Chinese lunar calendar holidays.
    The four day weekend rule seem to start in 2007 for these holidays.
    """
    friday_extras = [
        d + timedelta(1) for d in holidays
        if d.weekday() == THURSDAY and d.year > 2006
    ]

    monday_extras = [
        d - timedelta(1) for d in holidays
        if d.weekday() == TUESDAY and d.year > 2006
    ]

    return pd.to_datetime(friday_extras + monday_extras)


def taiwan_makeup_rule(holidays):
    """
    This function attempts to implement what seems to be the Taiwan holiday
    observance rule since 2013.

    Notes
    -----
    If a holiday falls on a Tuesday or Thursday, and extra holiday is observed
    on Monday/Friday to create a four day weekend.  If a holiday falls on a
    weekend, it is observed on the nearest workday (Monday/Friday).
    """
    # Create four day weekends
    mon = holidays[holidays.weekday == TUESDAY] - timedelta(1)
    fri = holidays[holidays.weekday == THURSDAY] + timedelta(1)

    # Nearest workday makeups
    mon_makeups = holidays[holidays.weekday == SUNDAY] + timedelta(1)
    fri_makeups = holidays[holidays.weekday == SATURDAY] - timedelta(1)

    # Only observe rule after 2013
    extras = pd.to_datetime(list(chain(mon, mon_makeups, fri, fri_makeups)))
    extras = extras[extras.year > 2013]

    return holidays.append(extras)


NewYearsDay = new_years_day(observance=taiwan_makeup_rule)

PeaceMemorialDay = Holiday(
    'Peace Memorial Day',
    month=2,
    day=28,
    observance=taiwan_makeup_rule,
)

WomenAndChildrensDay = Holiday(
    "Women and Children's Day",
    month=4,
    day=4,
    start_date='2011',
    observance=taiwan_makeup_rule,
)

LabourDay = european_labour_day(observance=sunday_to_monday)

NationalDay = Holiday(
    'National Day of the Republic of China',
    month=10,
    day=10,
    observance=taiwan_makeup_rule,
)

chinese_new_year = chinese_new_year_offset(chinese_lunar_new_year_dates)

chinese_new_years_eve = before_chinese_new_year_offset(
    chinese_new_year - timedelta(1),
)

chinese_new_years_eve_2 = before_chinese_new_year_offset(
    chinese_new_years_eve - timedelta(1),
)

chinese_new_year_2 = chinese_new_year_offset(
    chinese_new_year + timedelta(1),
)

chinese_new_year_3 = chinese_new_year_offset(
    chinese_new_year_2 + timedelta(1),
)

tomb_sweeping_day = manual_nearest_workday(qingming_festival_dates)

tomb_sweeping_day_extras = manual_extra_days(tomb_sweeping_day)

dragon_boat_festival = manual_nearest_workday(dragon_boat_festival_dates)

dragon_boat_festival_extras = manual_extra_days(dragon_boat_festival)

mid_autumn_festival = manual_nearest_workday(
    mid_autumn_festival_dates,
)

mid_autumn_festival_extras = manual_extra_days(mid_autumn_festival)

# Taiwan takes multiple days off before and after chinese new year,
# and sometimes it is unclear precisely which days will be holidays.
chinese_new_year_extras = pd.to_datetime([
    '2002-02-07',
    '2002-02-15',
    '2003-01-29',
    '2004-01-19',
    '2005-02-04',
    '2006-02-02',
    '2007-02-22',
    '2007-02-23',
    '2008-02-04',
    '2009-01-29',
    '2009-01-30',
    '2010-02-18',
    '2010-02-19',
    '2011-01-31',
    '2012-01-26',
    '2012-01-27',
    '2013-02-14',
    '2013-02-15',
    '2014-01-28',
    '2015-02-16',
    '2016-02-11',
    '2016-02-12',
    '2017-01-25',
    '2018-02-13',
    '2019-01-31',
    '2019-02-08',
    '2020-01-21',
    '2020-01-22',
])

# Some abnormal observances of regularly observed holidays.
extra_holidays = pd.to_datetime([
    '2020-04-02',  # Tomb Sweeping Day
    '2016-04-05',  # Tomb Sweeping Day
    '2012-12-31',  # New Year's Eve
    '2012-02-27',  # Peace Memorial Day
    '2009-01-02',  # New Year's Day
    '2006-10-09',  # National Day
    '2005-09-01',  # Bank Holiday
])

typhoons = pd.to_datetime([
    '2019-09-30',
    '2019-08-09',
    '2016-09-28',
    '2016-09-27',
    '2016-07-08',
    '2015-09-29',
    '2015-07-10',
    '2014-07-23',
    '2013-08-21',
    '2012-08-02',
    '2009-08-07',
    '2008-09-29',
    '2008-07-28',
    '2007-09-18',
    '2005-08-05',
    '2005-07-18',
    '2004-10-25',
    '2004-08-25',
    '2004-08-24',
    '2002-09-06',
])


class XTAIExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Taiwan Stock Exchange Corporation (XTAI).

    Open Time: 9:00 AM, CST
    Close Time: 1:30 PM, CST

    Regularly-Observed Holidays:
    - New Year's Day
    - Chinese New Year's Eve
    - Chinese New Year
    - Peace Memorial Day (Feb 28)
    - Women and Children's Day (Apr 4)
    - Tomb Sweeping Day (Lunar Calendar)
    - Labour Day (May 1)
    - Dragon Boat Festival (Lunar Calendar)
    - Mid-Autumn Festival (Lunar Calendar)
    - National Day of the Republic of China (Oct 10)


    Early Closes:
    - None
    """
    name = 'XTAI'

    tz = timezone('Asia/Taipei')

    open_times = (
        (None, time(9, 1)),
    )

    close_times = (
        (None, time(13, 30)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            PeaceMemorialDay,
            WomenAndChildrensDay,
            LabourDay,
            NationalDay,
        ])

    @property
    def adhoc_holidays(self):
        return list(chain(
            extra_holidays,
            typhoons,
            chinese_new_years_eve,
            chinese_new_years_eve_2,
            chinese_new_year,
            chinese_new_year_2,
            chinese_new_year_3,
            chinese_new_year_extras,
            tomb_sweeping_day,
            tomb_sweeping_day_extras,
            dragon_boat_festival,
            dragon_boat_festival_extras,
            mid_autumn_festival,
            mid_autumn_festival_extras,
        ))



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
import pandas as pd
from pandas.tseries.holiday import (
    DateOffset,
    EasterMonday,
    GoodFriday,
    Holiday,
    MO,
    previous_friday,
    weekend_to_monday,
)
from pytz import timezone

from .common_holidays import (
    new_years_day,
    christmas,
    weekend_christmas,
    boxing_day,
    weekend_boxing_day,
)
from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
)

# Regular Holidays
# ----------------
# New Year's Day
LSENewYearsDay = new_years_day(observance=weekend_to_monday)

# Early May bank holiday has two exceptions based on the 50th and 75th
# anniversary of VE-Day
# 1995-05-01 Early May bank holiday removed for VE-day 50th anniversary
# 2020-05-04 Early May bank holiday removed for VE-day 75th anniversary

# Early May bank holiday pre-1995
MayBank_pre_1995 = Holiday(
    "Early May Bank Holiday",
    month=5,
    offset=DateOffset(weekday=MO(1)),
    day=1,
    end_date=pd.Timestamp('1994-12-31'),
)

# Early May bank holiday post-1995 and pre-2020
MayBank_post_1995_pre_2020 = Holiday(
    "Early May Bank Holiday",
    month=5,
    offset=DateOffset(weekday=MO(1)),
    day=1,
    start_date=pd.Timestamp('1996-01-01'),
    end_date=pd.Timestamp('2019-12-31'),
)

# Early May bank holiday post 2020
MayBank_post_2020 = Holiday(
    "Early May Bank Holiday",
    month=5,
    offset=DateOffset(weekday=MO(1)),
    day=1,
    start_date=pd.Timestamp('2021-01-01')
)


# Spring bank holiday has two exceptions based on the Golden & Diamond Jubilee
# 2002-05-27 Spring bank holiday removed for Golden Jubilee
# 2012-05-28 Spring bank holiday removed for Diamond Jubilee

# Spring bank holiday
SpringBank_pre_2002 = Holiday(
    "Spring Bank Holiday",
    month=5,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
    end_date=pd.Timestamp('2001-12-31'),
)

SpringBank_post_2002_pre_2012 = Holiday(
    "Spring Bank Holiday",
    month=5,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
    start_date=pd.Timestamp('2003-01-01'),
    end_date=pd.Timestamp('2011-12-31'),
)

SpringBank_post_2012 = Holiday(
    "Spring Bank Holiday",
    month=5,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
    start_date=pd.Timestamp('2013-01-01'),
)

# Summer bank holiday
SummerBank = Holiday(
    "Summer Bank Holiday",
    month=8,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
)

Christmas = christmas()

WeekendChristmas = weekend_christmas()

BoxingDay = boxing_day()

WeekendBoxingDay = weekend_boxing_day()

# Early Closes
# ------------
# If Christmas Eve falls on a weekday, that day is a half day.
# If it falls on a weekend, the preceding Friday is a half day.
ChristmasEve = Holiday(
    'Christmas Eve Early Close',
    month=12,
    day=24,
    observance=previous_friday,
)
# New Year's eve (or the preceding Friday if it falls on a weekend)
# is a half day. Except for 1999-12-31, when the Queen declared a
# bank holiday.
NewYearsEvePre1999 = Holiday(
    "New Year's Eve Early Close",
    month=12,
    day=31,
    observance=previous_friday,
    end_date=pd.Timestamp('1999-01-01')
)
NewYearsEvePost2000 = Holiday(
    "New Year's Eve Early Close",
    month=12,
    day=31,
    observance=previous_friday,
    start_date=pd.Timestamp('2000-01-01')
)


class XLONExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the London Stock Exchange (XLON).

    Open Time: 8:00 AM, GMT
    Close Time: 4:30 PM, GMT

    Regularly-Observed Holidays:
    - New Years Day (observed on first business day on/after)
    - Good Friday
    - Easter Monday
    - Early May Bank Holiday (first Monday in May)
    - Spring Bank Holiday (last Monday in May)
    - Summer Bank Holiday (last Monday in May)
    - Christmas Day
    - Dec. 27th (if Christmas is on a weekend)
    - Boxing Day
    - Dec. 28th (if Boxing Day is on a weekend)

    Early Closes:
    - Christmas Eve
    - New Year's Eve
    """
    regular_early_close = time(12, 30)

    name = 'XLON'

    tz = timezone('Europe/London')

    open_times = (
        (None, time(8, 1)),
    )

    close_times = (
        (None, time(16, 30)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            LSENewYearsDay,
            GoodFriday,
            EasterMonday,
            MayBank_pre_1995,
            MayBank_post_1995_pre_2020,
            MayBank_post_2020,
            SpringBank_pre_2002,
            SpringBank_post_2002_pre_2012,
            SpringBank_post_2012,
            SummerBank,
            Christmas,
            WeekendChristmas,
            BoxingDay,
            WeekendBoxingDay
        ])

    @property
    def adhoc_holidays(self):
        return [
            # VE-Day Anniversary
            pd.Timestamp("1995-05-08", tz='UTC'),  # 50th Anniversary
            pd.Timestamp("2020-05-08", tz='UTC'),  # 75th Anniversary
            # Queen Elizabeth II Jubilees
            # Silver Jubilee
            pd.Timestamp("1977-06-07", tz='UTC'),
            # Golden Jubilee
            pd.Timestamp("2002-06-03", tz='UTC'),
            pd.Timestamp("2002-06-04", tz='UTC'),
            # Diamond Jubilee
            pd.Timestamp("2012-06-04", tz='UTC'),
            pd.Timestamp("2012-06-05", tz='UTC'),
            # Royal Weddings
            # Wedding Day of Princess Anne and Mark Phillips
            pd.Timestamp("1973-11-14", tz='UTC'),
            # Wedding Day of Prince Charles and Diana Spencer
            pd.Timestamp("1981-07-29", tz='UTC'),
            # Wedding Day of Prince William and Catherine Middleton
            pd.Timestamp("2011-04-29", tz='UTC'),
            # Miscellaneous
            # Eve of 3rd Millenium A.D.
            pd.Timestamp("1999-12-31", tz='UTC'),
        ]

    @property
    def special_closes(self):
        return [
            (self.regular_early_close, HolidayCalendar([
                ChristmasEve,
                NewYearsEvePre1999,
                NewYearsEvePost2000,
            ]))
        ]



"""
US Holidays

Many historical holidays were derived from the pdf at
etc/NYSE-Historical-Closings.pdf. Originally posted at
http://s3.amazonaws.com/armstrongeconomics-wp/2013/07/NYSE-Closings.pdf  # noqa

These were originally added in
https://github.com/rsheftel/pandas_market_calendars/pull/30
"""

from pandas import (
    Timestamp,
    DateOffset,
)
from pytz import UTC

from pandas.tseries.holiday import (
    Holiday,
    sunday_to_monday,
    nearest_workday,
)

from dateutil.relativedelta import (
    MO,
    TH,
    TU,
)
from pandas.tseries.offsets import Day

from .common_holidays import new_years_day
from .trading_calendar import (
    MONDAY,
    TUESDAY,
    WEDNESDAY,
    THURSDAY,
    FRIDAY,
)


def following_tuesday_every_four_years_observance(dt):
    return dt + DateOffset(years=(4 - (dt.year % 4)) % 4, weekday=TU(1))


# Holidays

# These have the same definition, but are used in different places because the
# NYSE closed at 2:00 PM on Christmas Eve until 1993.
ChristmasEveBefore1993 = Holiday(
    "Christmas Eve",
    month=12,
    day=24,
    end_date=Timestamp("1993-01-01"),
    # When Christmas is a Saturday, the 24th is a full holiday.
    days_of_week=(MONDAY, TUESDAY, WEDNESDAY, THURSDAY),
)
ChristmasEveInOrAfter1993 = Holiday(
    "Christmas Eve",
    month=12,
    day=24,
    start_date=Timestamp("1993-01-01"),
    # When Christmas is a Saturday, the 24th is a full holiday.
    days_of_week=(MONDAY, TUESDAY, WEDNESDAY, THURSDAY),
)
USNewYearsDay = new_years_day(
    # When Jan 1 is a Sunday, US markets observe the subsequent Monday.
    # When Jan 1 is a Saturday (as in 2005 and 2011), no holiday is observed.
    observance=sunday_to_monday
)
USMartinLutherKingJrAfter1998 = Holiday(
    "Dr. Martin Luther King Jr. Day",
    month=1,
    day=1,
    # The US markets didn't observe MLK day as a holiday until 1998.
    start_date=Timestamp("1998-01-01"),
    offset=DateOffset(weekday=MO(3)),
)
USLincolnsBirthDayBefore1954 = Holiday(
    "Lincoln's Birthday",
    month=2,
    day=12,
    start_date=Timestamp("1874-01-01"),
    end_date=Timestamp("1953-12-31"),
    observance=sunday_to_monday,
)
USWashingtonsBirthDayBefore1964 = Holiday(
    "Washington's Birthday",
    month=2,
    day=22,
    start_date=Timestamp("1880-01-01"),
    end_date=Timestamp("1963-12-31"),
    observance=sunday_to_monday,
)
USWashingtonsBirthDay1964to1970 = Holiday(
    "Washington's Birthday",
    month=2,
    day=22,
    start_date=Timestamp("1964-01-01"),
    end_date=Timestamp("1970-12-31"),
    observance=nearest_workday,
)
USPresidentsDay = Holiday(
    "President's Day",
    start_date=Timestamp("1971-01-01"),
    month=2,
    day=1,
    offset=DateOffset(weekday=MO(3)),
)

USThanksgivingDayBefore1939 = Holiday(
    "Thanksgiving Before 1939",
    start_date=Timestamp("1864-01-01"),
    end_date=Timestamp("1938-12-31"),
    month=11,
    day=30,
    offset=DateOffset(weekday=TH(-1)),
)

USThanksgivingDay1939to1941 = Holiday(
    "Thanksgiving 1939 to 1941",
    start_date=Timestamp("1939-01-01"),
    end_date=Timestamp("1941-12-31"),
    month=11,
    day=30,
    offset=DateOffset(weekday=TH(-2)),
)
USThanksgivingDay = Holiday(
    "Thanksgiving",
    start_date=Timestamp("1942-01-01"),
    month=11,
    day=1,
    offset=DateOffset(weekday=TH(4)),
)

USMemorialDayBefore1964 = Holiday(
    "Memorial Day",
    month=5,
    day=30,
    end_date=Timestamp("1963-12-31"),
    observance=sunday_to_monday,
)
USMemorialDay1964to1969 = Holiday(
    "Memorial Day",
    month=5,
    day=30,
    start_date=Timestamp("1964-01-01"),
    end_date=Timestamp("1969-12-31"),
    observance=nearest_workday,
)
USMemorialDay = Holiday(
    # NOTE: The definition for Memorial Day is incorrect as of pandas 0.16.0.
    # See https://github.com/pydata/pandas/issues/9760.
    "Memorial Day",
    month=5,
    day=25,
    start_date=Timestamp("1971-01-01"),
    offset=DateOffset(weekday=MO(1)),
)

USIndependenceDayBefore1954 = Holiday(
    "July 4th",
    month=7,
    day=4,
    end_date=Timestamp("1953-12-31"),
    observance=sunday_to_monday,
)
USIndependenceDay = Holiday(
    "July 4th",
    month=7,
    day=4,
    start_date=Timestamp("1954-01-01"),
    observance=nearest_workday,
)

USElectionDay1848to1967 = Holiday(
    "Election Day",
    month=11,
    day=2,
    start_date=Timestamp("1848-1-1"),
    end_date=Timestamp("1967-12-31"),
    offset=DateOffset(weekday=TU(1)),
)
USElectionDay1968to1980 = Holiday(
    "Election Day",
    month=11,
    day=2,
    start_date=Timestamp("1968-01-01"),
    end_date=Timestamp("1980-12-31"),
    observance=following_tuesday_every_four_years_observance,
)
USVeteransDay1934to1953 = Holiday(
    "Veteran Day",
    month=11,
    day=11,
    start_date=Timestamp("1934-1-1"),
    end_date=Timestamp("1953-12-31"),
    observance=sunday_to_monday,
)
USColumbusDayBefore1954 = Holiday(
    "Columbus Day",
    month=10,
    day=12,
    end_date=Timestamp("1953-12-31"),
    observance=sunday_to_monday,
)
ChristmasBefore1954 = Holiday(
    "Christmas",
    month=12,
    day=25,
    end_date=Timestamp("1953-12-31"),
    observance=sunday_to_monday,
)
Christmas = Holiday(
    "Christmas",
    month=12,
    day=25,
    start_date=Timestamp("1954-01-01"),
    observance=nearest_workday,
)

# Early Closes

MonTuesThursBeforeIndependenceDay = Holiday(
    # When July 4th is a Tuesday, Wednesday, or Friday, the previous day is a
    # half day.
    "Mondays, Tuesdays, and Thursdays Before Independence Day",
    month=7,
    day=3,
    days_of_week=(MONDAY, TUESDAY, THURSDAY),
    start_date=Timestamp("1995-01-01"),
)
FridayAfterIndependenceDayPre2013 = Holiday(
    # When July 4th is a Thursday, the next day is a half day prior to 2013.
    # Since 2013 the early close is on Wednesday and Friday is a full day
    "Fridays after Independence Day prior to 2013",
    month=7,
    day=5,
    days_of_week=(FRIDAY,),
    start_date=Timestamp("1995-01-01"),
    end_date=Timestamp("2013-01-01"),
)
WednesdayBeforeIndependenceDayPost2013 = Holiday(
    # When July 4th is a Thursday, the next day is a half day prior to 2013.
    # Since 2013 the early close is on Wednesday and Friday is a full day
    "Wednesdays Before Independence Day including and after 2013",
    month=7,
    day=3,
    days_of_week=(WEDNESDAY,),
    start_date=Timestamp("2013-01-01"),
)
USBlackFridayBefore1993 = Holiday(
    "Black Friday",
    month=11,
    day=1,
    # Black Friday was not observed until 1992.
    start_date=Timestamp("1992-01-01"),
    end_date=Timestamp("1993-01-01"),
    offset=[DateOffset(weekday=TH(4)), Day(1)],
)
USBlackFridayInOrAfter1993 = Holiday(
    "Black Friday",
    month=11,
    day=1,
    start_date=Timestamp("1993-01-01"),
    offset=[DateOffset(weekday=TH(4)), Day(1)],
)
BattleOfGettysburg = Holiday(
    # All of the floor traders in Chicago were sent to PA
    "Markets were closed during the battle of Gettysburg",
    month=7,
    day=(1, 2, 3),
    start_date=Timestamp("1863-07-01"),
    end_date=Timestamp("1863-07-03"),
)

# Adhoc and other closings
# use list for consistency in returning ad-hoc dates


November29BacklogRelief = [
    Timestamp("1929-11-01", tz="UTC"),
    Timestamp("1929-11-29", tz="UTC"),
]

March33BankHoliday = [
    Timestamp("1933-03-06", tz="UTC"),
    Timestamp("1933-03-07", tz="UTC"),
    Timestamp("1933-03-08", tz="UTC"),
    Timestamp("1933-03-09", tz="UTC"),
    Timestamp("1933-03-10", tz="UTC"),
    Timestamp("1933-03-13", tz="UTC"),
    Timestamp("1933-03-14", tz="UTC"),
]

August45VictoryOverJapan = [
    Timestamp("1945-08-15", tz="UTC"),
    Timestamp("1945-08-16", tz="UTC"),
]


ChristmasEvesAdhoc = [
    Timestamp("1945-12-24", tz="UTC"),
    Timestamp("1956-12-24", tz="UTC"),
]

DayAfterChristmasAdhoc = [Timestamp("1958-12-26", tz="UTC")]

DayBeforeDecorationAdhoc = [Timestamp("1961-05-29", tz="UTC")]

LincolnsBirthDayAdhoc = [Timestamp("1968-02-12", tz="UTC")]

PaperworkCrisis68 = [
    Timestamp("1968-06-12", tz="UTC"),
    Timestamp("1968-06-19", tz="UTC"),
    Timestamp("1968-06-26", tz="UTC"),
    Timestamp("1968-07-10", tz="UTC"),
    Timestamp("1968-07-17", tz="UTC"),
    Timestamp("1968-07-24", tz="UTC"),
    Timestamp("1968-07-31", tz="UTC"),
    Timestamp("1968-08-07", tz="UTC"),
    Timestamp("1968-08-14", tz="UTC"),
    Timestamp("1968-08-21", tz="UTC"),
    Timestamp("1968-08-28", tz="UTC"),
    Timestamp("1968-09-11", tz="UTC"),
    Timestamp("1968-09-18", tz="UTC"),
    Timestamp("1968-09-25", tz="UTC"),
    Timestamp("1968-10-02", tz="UTC"),
    Timestamp("1968-10-09", tz="UTC"),
    Timestamp("1968-10-16", tz="UTC"),
    Timestamp("1968-10-23", tz="UTC"),
    Timestamp("1968-10-30", tz="UTC"),
    Timestamp("1968-11-11", tz="UTC"),
    Timestamp("1968-11-20", tz="UTC"),
    Timestamp("1968-12-04", tz="UTC"),
    Timestamp("1968-12-11", tz="UTC"),
    Timestamp("1968-12-18", tz="UTC"),
    Timestamp("1968-12-25", tz="UTC"),
]

DayAfterIndependenceDayAdhoc = [Timestamp("1968-07-05", tz="UTC")]

WeatherSnowClosing = [Timestamp("1969-02-10", tz="UTC")]

FirstLunarLandingClosing = [Timestamp("1969-07-21", tz="UTC")]

NewYorkCityBlackout77 = [Timestamp("1977-07-14", tz="UTC")]


# http://en.wikipedia.org/wiki/Aftermath_of_the_September_11_attacks
September11Closings = [
    Timestamp("2001-09-11", tz=UTC),
    Timestamp("2001-09-12", tz=UTC),
    Timestamp("2001-09-13", tz=UTC),
    Timestamp("2001-09-14", tz=UTC),
]

# http://en.wikipedia.org/wiki/Hurricane_sandy
HurricaneSandyClosings = [
    Timestamp("2012-10-29", tz=UTC),
    Timestamp("2012-10-30", tz=UTC),
]

# add Hurricane Gloria closing
HurricaneGloriaClosing = [Timestamp("1985-09-27", tz=UTC)]


# National Days of Mourning
# - President John F. Kennedy - November 25, 1963
# - Martin Luther King - April 9, 1968
# - President Dwight D. Eisenhower - March 31, 1969
# - President Harry S. Truman - December 28, 1972
# - President Lyndon B. Johnson - January 25, 1973
# - President Richard Nixon - April 27, 1994
# - President Ronald W. Reagan - June 11, 2004
# - President Gerald R. Ford - Jan 2, 2007
# - President George H.W. Bush - Dec 5, 2018
# added Truman and Johnson to go back to 1970
# http://s3.amazonaws.com/armstrongeconomics-wp/2013/07/NYSE-Closings.pdf
USNationalDaysofMourning = [
    Timestamp("1963-11-25", tz="UTC"),
    Timestamp("1968-04-09", tz="UTC"),
    Timestamp("1969-03-31", tz="UTC"),
    Timestamp("1972-12-28", tz=UTC),
    Timestamp("1973-01-25", tz=UTC),
    Timestamp("1994-04-27", tz=UTC),
    Timestamp("2004-06-11", tz=UTC),
    Timestamp("2007-01-02", tz=UTC),
    Timestamp("2018-12-05", tz=UTC),
]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from pandas.tseries.holiday import (
    DateOffset,
    Easter,
    GoodFriday,
    Holiday,
    MO,
    previous_friday,
)
from pandas.tseries.offsets import Day
from pytz import timezone

from .common_holidays import (
    epiphany,
    immaculate_conception,
    maundy_thursday,
    new_years_day,
    new_years_eve,
    european_labour_day,
    christmas,
)
from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
)

# Used for "first Monday on or after date"
next_monday_offset = DateOffset(weekday=MO(1))

NewYearsDay = new_years_day()

Epiphany = epiphany(observance=next_monday_offset)

StJosephsDay = Holiday(
    "St. Joseph's Day (next Monday)",
    month=3,
    day=19,
    offset=next_monday_offset,
)

MaundyThursday = maundy_thursday()

LabourDay = european_labour_day()

MondayAfterAscensionDay = Holiday(
    "Monday After Ascension Day",
    month=1,
    day=1,
    offset=[Easter(), Day(43)],
)

MondayAfterCorpusChristi = Holiday(
    "Monday After Corpus Christi",
    month=1,
    day=1,
    offset=[Easter(), Day(64)],
)

MondayAfterSacredHeart = Holiday(
    "Monday After Sacred Heart",
    month=1,
    day=1,
    offset=[Easter(), Day(71)],
)

StPeterAndStPaulDay = Holiday(
    "St. Peter and St. Paul Day",
    month=6,
    day=29,
    offset=next_monday_offset
)

ColombiaIndependenceDay = Holiday(
    "Colombian Independence Day",
    month=7,
    day=20,
)

BattleOfBoyaca = Holiday(
    "Battle of Boyaca",
    month=8,
    day=7,
)

AssumptionDay = Holiday(
    "Assumption Day (next Monday)",
    month=8,
    day=15,
    offset=next_monday_offset,
)

DiaDeLaRaza = Holiday(
    'Dia de la Raza',
    month=10,
    day=12,
    offset=next_monday_offset,
)

AllSaintsDay = Holiday(
    "All Saint's Day (next Monday)",
    month=11,
    day=1,
    offset=next_monday_offset,
)

CartagenaIndependenceDay = Holiday(
    "Cartagena Independence Day",
    month=11,
    day=11,
    observance=next_monday_offset,
)

ImmaculateConception = immaculate_conception()

ChristmasDay = christmas()

LastTradingDay = new_years_eve(observance=previous_friday)


class XBOGExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Colombia Stock Exchange (XBOG).

    Open Time: 9:30 AM, EST
    Close Time: 4:00 PM, EST

    Regularly-Observed Holidays:
    - New Years Day
    - Epiphany
    - St. Joseph's Day (first Monday on/after March 19)
    - Maundy Thursday (Thursday before Easter)
    - Good Friday
    - Labour Day (May 1)
    - Ascension Day (first Monday >= 39 days after Easter Sunday)
    - Corpus Christi (first Monday >= 60 days after Easter Sunday)
    - Sacred Heart (first Monday >= 68 days after Easter Sunday)
    - Saint Peter and Saint Paul Day (first Monday on or after June 29)
    - Independence Day (July 20)
    - Battle of Boyaca (August 7)
    - Assumption Day (first Monday on or after August 15)
    - Columbus Day (first Monday on or after October 12)
    - All Saint's Day (first Monday on or after November 1)
    - Cartagena Independence Day (November 11)
    - Immaculate Conception (December 8)
    - Christmas Day

    Early Closes:
    - None
    """
    name = 'XBOG'

    # Though Bogota uses Colombia Standard Time, XBOG uses
    # US Eastern for trading times
    tz = timezone('America/New_York')

    open_times = (
        (None, time(9, 31)),
    )

    close_times = (
        (None, time(16)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            Epiphany,
            StJosephsDay,
            MaundyThursday,
            GoodFriday,
            LabourDay,
            MondayAfterAscensionDay,
            MondayAfterCorpusChristi,
            MondayAfterSacredHeart,
            StPeterAndStPaulDay,
            ColombiaIndependenceDay,
            BattleOfBoyaca,
            AssumptionDay,
            DiaDeLaRaza,
            AllSaintsDay,
            CartagenaIndependenceDay,
            ImmaculateConception,
            ChristmasDay,
            LastTradingDay,
        ])



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

import pandas as pd
from pandas.tseries.holiday import GoodFriday, Holiday, MO
from pandas.tseries.offsets import DateOffset
from pytz import timezone, UTC

from .common_holidays import (
    new_years_day,
    maundy_thursday,
    european_labour_day,
    christmas,
)
from .trading_calendar import HolidayCalendar, TradingCalendar


NewYearsDay = new_years_day()

# Starting in 2007 Constitution Day is observed on the first Monday of Feb.
ConstitutionDay = Holiday('Constitution Day', month=2, day=5, end_date='2007')
ConstitutionDayObserved = Holiday(
    'Constitution Day Observed',
    month=2,
    day=1,
    offset=DateOffset(weekday=MO(1)),
    start_date='2007',
)

# Starting in 2007 Juarez's Birthay is observed on the third Monday of March.
JuarezBirthday = Holiday("Juarez's Birthday", month=3, day=21, end_date='2007')
JuarezBirthdayObserved = Holiday(
    "Juarez's Birthday Observed",
    month=3,
    day=1,
    offset=DateOffset(weekday=MO(3)),
    start_date='2007',
)

MaundyThursday = maundy_thursday()

LabourDay = european_labour_day()

IndependenceDay = Holiday('Independence Day', month=9, day=16)

AllSoulsDay = Holiday("All Souls' Day", month=11, day=2, start_date='2006')

# Starting in 2007 Revolution Day is observed on the third Monday of November.
RevolutionDay = Holiday('Revolution Day', month=11, day=20, end_date='2007')
RevolutionDayObserved = Holiday(
    'Revolution Day Observed',
    month=11,
    day=1,
    offset=DateOffset(weekday=MO(3)),
    start_date='2007',
)

BankingHoliday = Holiday('Banking Holiday', month=12, day=12)

Christmas = christmas()


class XMEXExchangeCalendar(TradingCalendar):
    """
    Calendar for the Mexican Stock Exchange (Bolsa Mexicana de Valores) in
    Mexico City.

    Open Time: 8:30 AM, CT (Central Time)
    Close Time: 3:00 PM, CT (Central Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Constitution Day
      - Juarez's Birthday
      - Maundy Thursday
      - Good Friday
      - Labour Day
      - Independence Day
      - All Souls' Day
      - Mexican Revolution
      - Banking Holiday
      - Christmas Day

    Holidays No Longer Observed:
      - None

    Early Closes:
      - None
    """
    name = 'XMEX'
    tz = timezone('America/Mexico_City')

    open_times = (
        (None, time(8, 31)),
    )
    close_times = (
        (None, time(15)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            ConstitutionDay,
            ConstitutionDayObserved,
            JuarezBirthday,
            JuarezBirthdayObserved,
            MaundyThursday,
            GoodFriday,
            LabourDay,
            IndependenceDay,
            AllSoulsDay,
            RevolutionDay,
            RevolutionDayObserved,
            BankingHoliday,
            Christmas,
        ])

    @property
    def adhoc_holidays(self):
        return [
            # Bicentennial Celebration.
            pd.Timestamp('2010-09-17', tz=UTC),
        ]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta
from itertools import chain
import pandas as pd
from pandas.tseries.holiday import (
    Easter,
    GoodFriday,
    Holiday,
    MO,
)
from pandas.tseries.offsets import Day
from pytz import timezone

from .common_holidays import (
    european_labour_day,
    new_years_day,
    immaculate_conception,
    maundy_thursday,
    christmas,
    christmas_eve,
)
from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
    MONDAY,
    TUESDAY,
    WEDNESDAY,
    WEEKDAYS,
)


def nearest_monday(dt):
    """
    If a holiday falls on Tuesday or Wednesday, it is observed on the
    previous Monday.  If it falls on Thursday, Friday or a weekend,
    it is observed on the following Monday.
    """
    day = dt.weekday()

    if day in (MONDAY, TUESDAY, WEDNESDAY):
        return dt - timedelta(day)

    while dt.weekday() != MONDAY:
        dt += timedelta(1)

    return dt


def cultural_diversity_observance(holidays):
    """
    Apply the nearest monday rule, and also exclude 2012 (Day of
    Respect for Cultural Diversity breaks the nearest monday rule
    in 2012).
    """
    holidays = pd.to_datetime(
        holidays.map(nearest_monday)
    )
    return holidays[holidays.year != 2012]


def not_2018(holidays):
    """
    Exclude the year 2018 from the holiday list.
    """
    return holidays[holidays.year != 2018]


NewYearsDay = new_years_day()

CarnivalMonday = Holiday(
    'Carnival Monday',
    month=1,
    day=1,
    start_date='2011',
    offset=[Easter(), -Day(48)],
)

CarnivalTuesday = Holiday(
    'Carnival Tuesday',
    month=1,
    day=1,
    start_date='2011',
    offset=[Easter(), -Day(47)],
)

TruthAndJusticeMemorialDay = Holiday(
    'Truth and Justice Memorial Day',
    month=3,
    day=24,
    start_date='2005',
)

MalvinasDayTo2004 = Holiday(
    'Malvinas Day (up to 2004)',
    month=3,
    day=31,
    end_date='2005',
    offset=pd.DateOffset(weekday=MO(1)),
)

MalvinasDayFrom2005 = Holiday(
    'Malvinas Day (2005-present)',
    month=4,
    day=2,
    start_date='2005'
)

MaundyThursday = maundy_thursday()

LabourDay = european_labour_day()

MayDayRevolution = Holiday(
    'May Day Revolution',
    month=5,
    day=25,
)

MartinMiguelDeGuemesDay = Holiday(
    'Martin Miguel de Guemes Day',
    month=6,
    day=17,
    start_date='2016',
)

NationalFlagDayTo2010 = Holiday(
    'National Flag Day (up to 2010)',
    month=6,
    day=1,
    end_date='2011',
    offset=pd.DateOffset(weekday=MO(3)),
)

NationalFlagDayFrom2011 = Holiday(
    'National Flag Day (2011-present)',
    month=6,
    day=20,
    start_date='2011',
)

IndependenceDay = Holiday(
    'Independence Day',
    month=7,
    day=9,
)

SanMartinsDayTo2010 = Holiday(
    "San Martin's Day",
    month=8,
    day=1,
    end_date='2011',
    offset=pd.DateOffset(weekday=MO(3)),
)

SanMartinsDayFrom2012 = Holiday(
    "San Martin's Day",
    month=8,
    day=1,
    start_date='2012',
    offset=pd.DateOffset(weekday=MO(3)),
)

DayOfRespectForCulturalDiversity = Holiday(
    "Day of Respect for Cultural Diversity",
    month=10,
    day=12,
    observance=cultural_diversity_observance,
)

BankHoliday = Holiday(
    'Bank Holiday',
    month=11,
    day=6,
    end_date='2018',
)

DayOfNationalSovereigntyTo2014 = Holiday(
    'Day of National Sovereignty (2010-2014)',
    month=11,
    day=1,
    start_date='2010',
    end_date='2015',
    offset=pd.DateOffset(weekday=MO(4)),
)

DayOfNationalSovereigntyFrom2017 = Holiday(
    'Day of National Sovereignty (2017-present)',
    month=11,
    day=17,
    start_date='2017',
    offset=pd.DateOffset(weekday=MO(1)),
)

ImmaculateConception = immaculate_conception()

ChristmasEve = christmas_eve(
    days_of_week=WEEKDAYS,
    observance=not_2018,
)

ChristmasDay = christmas()

NewYearsEve = Holiday(
    "New Year's Eve",
    month=12,
    day=31,
    days_of_week=WEEKDAYS,
    observance=not_2018,
)


class XBUEExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Buenos Aires Stock Exchange (XBUE).

    Open Time: 11:00 AM, AST
    Close Time: 5:00 PM, AST

    Regularly-Observed Holidays:
    - New Year's Day
    - Carnival Monday (48 days before Easter)
    - Carnival Tuesday (47 days before Easter)
    - Truth and Justice Memorial Day (Mar 24)
    - Malvinas Day (until 2004 first Monday after Mar 31, 2005-present Apr 2)
    - Good Friday
    - Labour Day
    - May Day Revolution (May 25)
    - Martin Miguel de Guemes Day (Jun 17, 2016-present)
    - National Flag Day (until 2010 3rd Monday in Jun, 2011-present Jun 20)
    - Independence Day (Jul 9)
    - San Martin's Day (3rd Monday in Aug)
    - Day of Respect for Cultural Diversity (2nd Monday in Oct)
    - Bank Holiday (Nov 6)
    - Day of National Sovereignty (4th Monday in Nov)
    - Immaculate Conception
    - Christmas Eve (2005-present)
    - Christmas Day
    - Last Trading Day of Year (2005-present)


    Early Closes:
    Note - we are unsure if these are early closes or holidays
    - Christmas Eve
    - New Year's Eve
    """
    name = 'XBUE'

    tz = timezone('America/Argentina/Buenos_Aires')

    open_times = (
        (None, time(11, 1)),
    )

    close_times = (
        (None, time(17, 00)),
    )

    # TODO: Verify early close time
    regular_early_close = time(13)

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            CarnivalMonday,
            CarnivalTuesday,
            TruthAndJusticeMemorialDay,
            MalvinasDayTo2004,
            MalvinasDayFrom2005,
            MaundyThursday,
            GoodFriday,
            LabourDay,
            MayDayRevolution,
            MartinMiguelDeGuemesDay,
            NationalFlagDayTo2010,
            NationalFlagDayFrom2011,
            IndependenceDay,
            SanMartinsDayTo2010,
            SanMartinsDayFrom2012,
            DayOfRespectForCulturalDiversity,
            BankHoliday,
            DayOfNationalSovereigntyTo2014,
            DayOfNationalSovereigntyFrom2017,
            ImmaculateConception,
            ChristmasDay,
        ])

    @property
    def adhoc_holidays(self):
        misc_adhocs = [
            '2009-07-10',
            '2010-05-24',
            '2010-10-27',
            '2012-01-02',
            '2012-02-27',
            '2012-09-24',
            '2013-01-31',
            '2013-02-20',
            '2018-11-30',
        ]

        # There was a string of market closures in January, 2002
        market_closures_2002_jan = [
            '2002-01-{:02d}'.format(day) for day in range(7, 17)
        ]

        # There was a string of market closures in April, 2002
        market_closures_2002_apr = [
            '2002-04-{:02d}'.format(day) for day in range(22, 27)
        ]

        # There are occasionally "bridge days" that make long weekends when
        # a holiday falls on a Tuesday/Thursday, but these days are chosen
        # inconsistently.
        bridge_days = [
            '2011-03-25',
            '2011-12-09',
            '2012-04-30',
            '2013-04-01',
            '2013-06-21',
            '2014-05-02',
            '2014-12-26',
            '2015-03-23',
            '2015-12-07',
            '2016-07-08',
            '2016-12-09',
            '2018-04-30',
            '2018-12-24',
            '2018-12-31',
            '2019-07-08',
            '2019-08-19',
            '2019-10-14',
        ]

        irregular_observances = [
            '2012-10-08',  # Cultural Diversity Day
            '2011-08-22',  # San Martin's Day
            '2015-11-27',  # Day of National Sovereignty
            '2016-11-28',  # Day of National Sovereignty
        ]

        return list(chain(
            misc_adhocs,
            market_closures_2002_jan,
            market_closures_2002_apr,
            bridge_days,
            irregular_observances,
        ))

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar([
                    ChristmasEve,
                    NewYearsEve,
                ]),
            ),
        ]



from __future__ import print_function

import sys


months = [
    'January',
    'February',
    'March',
    'April',
    'May',
    'June',
    'July',
    'August',
    'September',
    'October',
    'November',
    'December',
]


def error(msg):
    print(msg, file=sys.stderr)
    exit(-1)


def _render_month(calendar, year, month, print_year):
    import pandas as pd

    if sys.version_info[0] == 2:
        import StringIO

        out = StringIO.StringIO()
    else:
        import io

        out = io.StringIO()

    start = '{year}-{month}'.format(year=year, month=month)
    if month == 12:
        end = '{year}-{month}'.format(year=year + 1, month=1)
    else:
        end = '{year}-{month}'.format(year=year, month=month + 1)

    days = pd.date_range(start, end, closed='left')

    title = months[month - 1]
    if print_year:
        title += ' {year}'.format(year=year)

    print('{title:^28}'.format(title=title).rstrip(), file=out)
    print(' Su  Mo  Tu  We  Th  Fr  Sa', file=out)
    print(
        ' ' * (4 * ((days[0].weekday() + 1) % 7)),
        end='',
        file=out,
    )

    for d in days:
        if d.weekday() == 6:
            print('', file=out)

        if calendar.is_session(d):
            a = b = ' '
        else:
            a = '['
            b = ']'

        print(
            '{a}{d.day:>2}{b}'.format(a=a, d=d, b=b),
            end='',
            file=out,
        )

    print('', file=out)
    return out.getvalue()


def _concat_lines(strings, width):
    as_lines = [string.splitlines() for string in strings]
    max_lines = max(len(lines) for lines in as_lines)
    for lines in as_lines:
        missing_lines = max_lines - len(lines)
        if missing_lines:
            lines.extend([' ' * width] * missing_lines)

    rows = []
    for row_parts in zip(*as_lines):
        row_parts = list(row_parts)
        for n, row_part in enumerate(row_parts):
            missing_space = width - len(row_part)
            if missing_space:
                row_parts[n] = row_part + ' ' * missing_space

        rows.append('   '.join(row_parts))

    return '\n'.join(row.rstrip() for row in rows)


def _int_arg(v, name):
    try:
        return int(v)
    except ValueError:
        error('%s must be an integer, got: %r' % (name, v))


def parse_args(argv):
    usage = 'usage: %s CALENDAR [[[DAY] MONTH] YEAR]' % argv[0]

    if len(argv) == 1 or '--help' in argv or '-h' in argv:
        error(usage)

    if len(argv) > 1:
        from trading_calendars import get_calendar

        try:
            calendar = get_calendar(argv[1])
        except Exception as e:
            error(str(e))

    if len(argv) == 2:
        import datetime

        now = datetime.datetime.now()
        year = now.year
        month = now.month
    elif len(argv) == 3:
        year = _int_arg(argv[2], 'YEAR')
        month = None
    elif len(argv) == 4:
        month = _int_arg(argv[2], 'MONTH')
        year = _int_arg(argv[3], 'YEAR')
    else:
        error(usage)

    return calendar, year, month


def main(argv=None):
    """Print a unix-cal like calendar but indicate which days are trading
    sessions.
    """
    if argv is None:
        argv = sys.argv
    calendar, year, month = parse_args(argv)

    if month is not None:
        print(_render_month(calendar, year, month, print_year=True))
    else:
        month_strings = [
            [
                _render_month(
                    calendar,
                    year,
                    row * 3 + column + 1,
                    print_year=False,
                )
                for column in range(3)
            ]
            for row in range(4)
        ]
        print('{year:^88}\n'.format(year=year).rstrip())
        print('\n\n'.join(_concat_lines(cs, 28) for cs in month_strings))


if __name__ == '__main__':
    main()



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

from pytz import timezone

from .trading_calendar import HolidayCalendar, TradingCalendar
from .xbkk_holidays import (
    NewYearsDay,
    ChakriMemorialDay,
    SongkranFestival1,
    SongkranFestival2,
    SongkranFestival3,
    LabourDay,
    CoronationDay2016AndBefore,
    CoronationDay2019AndAfter,
    HMQueensBirthday,
    HMKingsBirthday,
    HMQueenMothersBirthday,
    HalfYearHoliday,
    ThePassingOfKingBhumibol,
    ChulalongkornDay,
    KingBhumibolsBirthday,
    ThailandConstitutionDay,
    NewYearsEve,
    makha_bucha,
    vesak,
    asanha_bucha,
    new_years_bridge_days,
    asanha_bucha_bridge_days,
    queens_birthday_bridge_days,
    coronation_bridge_days,
    vesak_bridge_days,
    misc_adhoc,
)


class XBKKExchangeCalendar(TradingCalendar):
    """
    Calendar for the Stock Exchange of Thailand in Bangkok.

    Open Time: 10:00 AM, Indochina Time (ICT)
    Close Time: 4:30 PM, Indochina Time (ICT)

    Regularly-Observed Holidays:
      - New Year's Day
      - Makha Bucha
      - Chakri Memorial Day
      - Songkran Festival
      - Labour Day
      - Coronation Day
      - Vesak
      - Her Majesty The Queen's Birthday
      - Asanha Bucha
      - His Majesty The King's Birthday
      - Her Majesty The Queen Mother's Birthday
      - The Passing of King Bhumibol
      - Chulalongkorn Day
      - King Bhumibol's Birthday
      - Thailand Constitution Day
      - New Year's Eve

    Early Closes:
      - None
    """
    name = 'XBKK'

    tz = timezone('Asia/Bangkok')

    open_times = (
        (None, time(10, 1)),
    )

    close_times = (
        (None, time(16, 30)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            ChakriMemorialDay,
            SongkranFestival1,
            SongkranFestival2,
            SongkranFestival3,
            LabourDay,
            CoronationDay2016AndBefore,
            CoronationDay2019AndAfter,
            HMQueensBirthday,
            HMKingsBirthday,
            HMQueenMothersBirthday,
            HalfYearHoliday,
            ThePassingOfKingBhumibol,
            ChulalongkornDay,
            KingBhumibolsBirthday,
            ThailandConstitutionDay,
            NewYearsEve,
        ])

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                makha_bucha,
                vesak,
                asanha_bucha,
                new_years_bridge_days,
                asanha_bucha_bridge_days,
                queens_birthday_bridge_days,
                coronation_bridge_days,
                vesak_bridge_days,
                misc_adhoc,
            ),
        )



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import Holiday, weekend_to_monday
from pytz import timezone, UTC

from .common_holidays import new_years_day, new_years_eve, european_labour_day
from .trading_calendar import HolidayCalendar, TradingCalendar, WEEKDAYS


def new_years_eve_observance(holidays):
    # For some reason New Year's Eve was not a holiday these years.
    holidays = holidays[
        (holidays.year != 2008) &
        (holidays.year != 2009)
    ]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


def new_years_holiday_observance(holidays):
    # New Year's Holiday did not follow the next-non-holiday rule in 2016.
    holidays = holidays[(holidays.year != 2016)]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


def orthodox_christmas_observance(holidays):
    # Orthodox Christmas did not follow the next-non-holiday rule these years.
    holidays = holidays[
        (holidays.year != 2012) &
        (holidays.year != 2017)
    ]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


def defender_of_fatherland_observance(holidays):
    # Defender of the Fatherland Day did not follow the next-non-holiday rule
    # these years.
    holidays = holidays[
        (holidays.year != 2013) &
        (holidays.year != 2014) &
        (holidays.year != 2019)
    ]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


NewYearsDay = new_years_day(observance=weekend_to_monday)
NewYearsHoliday = Holiday(
    "New Year's Holiday",
    month=1,
    day=2,
    observance=new_years_holiday_observance,
)
NewYearsHoliday2 = Holiday(
    "New Year's Holiday",
    month=1,
    day=3,
    start_date='2005',
    end_date='2012',
)
NewYearsHoliday3 = Holiday(
    "New Year's Holiday",
    month=1,
    day=4,
    start_date='2005',
    end_date='2012',
)
NewYearsHoliday4 = Holiday(
    "New Year's Holiday",
    month=1,
    day=5,
    start_date='2005',
    end_date='2012',
)
NewYearsHoliday5 = Holiday(
    "New Year's Holiday",
    month=1,
    day=6,
    start_date='2005',
    end_date='2012',
)

OrthodoxChristmas = Holiday(
    'Orthodox Christmas',
    month=1,
    day=7,
    observance=orthodox_christmas_observance,
)

DefenderOfTheFatherlandDay = Holiday(
    'Defender of the Fatherland Day',
    month=2,
    day=23,
    observance=defender_of_fatherland_observance,
)

WomensDay = Holiday(
    "Women's Day",
    month=3,
    day=8,
    observance=weekend_to_monday,
)

LabourDay = european_labour_day(observance=weekend_to_monday)

VictoryDay = Holiday(
    'Victory Day',
    month=5,
    day=9,
    observance=weekend_to_monday,
)

DayOfRussia = Holiday(
    'Day of Russia',
    month=6,
    day=12,
    observance=weekend_to_monday,
)

UnityDay = Holiday(
    'Unity Day',
    month=11,
    day=4,
    observance=weekend_to_monday,
    start_date='2005',
)

NewYearsEve = new_years_eve(
    observance=new_years_eve_observance,
    days_of_week=WEEKDAYS,
)


# Adhoc Holidays
# --------------

# All of the following "extensions" are bridge holidays, meaning they are
# either a Monday or Friday that is made into a holiday to fill in the gap
# between a Tuesday or Thursday holiday, respectively. Unfortunately having
# these bridge days is not consistently the rule, so they are treated as adhoc.
# This means that in the future there may be manual additions needed here.
new_years_extensions = [
    pd.Timestamp('2003-01-03', tz=UTC),
    pd.Timestamp('2013-01-03', tz=UTC),
    pd.Timestamp('2013-01-04', tz=UTC),
    pd.Timestamp('2014-01-03', tz=UTC),
]

orthodox_christmas_extensions = [
    pd.Timestamp('2003-01-06', tz=UTC),
    pd.Timestamp('2005-01-10', tz=UTC),
    pd.Timestamp('2008-01-08', tz=UTC),
    pd.Timestamp('2009-01-08', tz=UTC),
    pd.Timestamp('2009-01-09', tz=UTC),
    pd.Timestamp('2010-01-08', tz=UTC),
    pd.Timestamp('2011-01-10', tz=UTC),
    pd.Timestamp('2016-01-08', tz=UTC),
]

defender_of_the_fatherland_extensions = [
    pd.Timestamp('2006-02-24', tz=UTC),
    pd.Timestamp('2010-02-22', tz=UTC),
]

womens_day_extensions = [
    pd.Timestamp('2005-03-07', tz=UTC),
    pd.Timestamp('2011-03-07', tz=UTC),
    pd.Timestamp('2012-03-09', tz=UTC),
]

labour_day_extensions = [
    pd.Timestamp('2002-05-02', tz=UTC),
    pd.Timestamp('2002-05-03', tz=UTC),
    pd.Timestamp('2003-05-02', tz=UTC),
    pd.Timestamp('2004-05-04', tz=UTC),
    pd.Timestamp('2007-04-30', tz=UTC),
    pd.Timestamp('2008-05-02', tz=UTC),
    pd.Timestamp('2012-04-30', tz=UTC),
    pd.Timestamp('2015-05-04', tz=UTC),
    pd.Timestamp('2016-05-03', tz=UTC),
]

victory_day_extensions = [
    pd.Timestamp('2002-05-10', tz=UTC),
    pd.Timestamp('2005-05-10', tz=UTC),
    pd.Timestamp('2006-05-08', tz=UTC),
    pd.Timestamp('2017-05-08', tz=UTC),
]

day_of_russia_extensions = [
    pd.Timestamp('2003-06-13', tz=UTC),
    pd.Timestamp('2007-06-11', tz=UTC),
    pd.Timestamp('2008-06-13', tz=UTC),
    pd.Timestamp('2012-06-11', tz=UTC),
    pd.Timestamp('2014-06-13', tz=UTC),
]

unity_day_extensions = [
    pd.Timestamp('2008-11-03', tz=UTC),
    pd.Timestamp('2010-11-05', tz=UTC),
]

misc_adhoc = [
    # Exchange Holidays.
    pd.Timestamp('2002-11-07', tz=UTC),
    pd.Timestamp('2002-11-08', tz=UTC),
    pd.Timestamp('2002-12-12', tz=UTC),
    pd.Timestamp('2002-12-13', tz=UTC),
    pd.Timestamp('2003-11-07', tz=UTC),
    pd.Timestamp('2003-12-12', tz=UTC),
    pd.Timestamp('2004-11-08', tz=UTC),
    pd.Timestamp('2004-12-13', tz=UTC),
    pd.Timestamp('2008-09-18', tz=UTC),
    # Trading Suspended.
    pd.Timestamp('2008-10-10', tz=UTC),
    pd.Timestamp('2008-10-27', tz=UTC),
]


class XMOSExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Moscow Stock Exchange.

    Open Time: 10:00 AM, MSK (Moscow Standard Time)
    Close Time: 6:45 PM, MSK (Moscow Standard Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - New Year's Holiday
      - Orthodox Christmas Day
      - Defender of the Fatherland Day
      - Women's Day
      - Spring and Labour Day
      - Victory Day
      - Day of Russia
      - Unity Day
      - New Year's Eve

    Holidays No Longer Observed:
      - New Year's Holiday Week

    Early Closes:
      - None
    """
    name = 'XMOS'

    tz = timezone('Europe/Moscow')

    open_times = (
        (None, time(10, 1)),
    )

    close_times = (
        (None, time(18, 45)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            NewYearsHoliday,
            NewYearsHoliday2,
            NewYearsHoliday3,
            NewYearsHoliday4,
            NewYearsHoliday5,
            OrthodoxChristmas,
            DefenderOfTheFatherlandDay,
            WomensDay,
            LabourDay,
            VictoryDay,
            DayOfRussia,
            UnityDay,
            NewYearsEve,
        ])

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                new_years_extensions,
                orthodox_christmas_extensions,
                defender_of_the_fatherland_extensions,
                womens_day_extensions,
                labour_day_extensions,
                victory_day_extensions,
                day_of_russia_extensions,
                unity_day_extensions,
                misc_adhoc,
            ),
        )



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from pandas.tseries.offsets import Day
from pandas import Timestamp
from pandas.tseries.holiday import (
    Easter,
    GoodFriday,
    Holiday,
    sunday_to_monday,
)
from pytz import timezone, UTC

from .common_holidays import new_years_day
from .trading_calendar import HolidayCalendar, TradingCalendar


class XJSEExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Johannesburg Stock Exchange (XJSE).

    Open Time: 09:00, SAST
    Close Time: 17:00, SAST

    Regularly-Observed Holidays:
    - New Year's Day (Jan 1)
    - Human Rights Day (Mar 31)
    - Good Friday (Friday before Easter)
    - Family Day (Monday after Easter)
    - Freedom Day (Apr 27)
    - Workers' Day (a.k.a. Labour Day, May 1)
    - Youth Day (Jun 16)
    - National Women's Day (Aug 19)
    - Heritage Day (Sep 24)
    - Day of Reconciliation (Dec 19)
    - Christmas (Dec 25)
    - Day of Goodwill (Dec 26)

    If a holiday falls on a Saturday, it is not observed.
    If a holiday falls on a Sunday, it is observed on Monday.

    If two holiday observations occur on the same trading day, the
    following trading day may or may not be declared a holiday: if so,
    it should be added to ``adhoc_holidays``.
    """
    name = 'XJSE'

    tz = timezone('Africa/Johannesburg')

    open_times = (
        (None, time(9, 1)),
    )

    close_times = (
        (None, time(17, 00)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            new_years_day(
                observance=sunday_to_monday,
            ),
            Holiday(
                "Human Rights Day",
                month=3,
                day=21,
                observance=sunday_to_monday,
            ),
            GoodFriday,
            Holiday(
                "Family Day",
                month=1,
                day=1,
                offset=[Easter(), Day(1)],
            ),
            Holiday(
                "Freedom Day",
                month=4,
                day=27,
                observance=sunday_to_monday,
            ),
            Holiday(
                "Workers' Day",
                month=5,
                day=1,
                observance=sunday_to_monday,
            ),
            Holiday(
                "Youth Day",
                month=6,
                day=16,
                observance=sunday_to_monday,
            ),
            Holiday(
                "National Women's Day",
                month=8,
                day=9,
                observance=sunday_to_monday,
            ),
            Holiday(
                "Heritage Day",
                month=9,
                day=24,
                observance=sunday_to_monday,
            ),
            Holiday(
                "Day of Reconciliation",
                month=12,
                day=16,
                observance=sunday_to_monday,
            ),
            Holiday(
                "Christmas",
                month=12,
                day=25,
                observance=sunday_to_monday,
            ),
            Holiday(
                "Day of Goodwill",
                month=12,
                day=26,
                observance=sunday_to_monday,
            ),
        ])

    @property
    def adhoc_holidays(self):
        return [Timestamp(date, tz=UTC) for date in [
            # Election holidays
            '2004-04-14',
            '2006-03-01',
            '2009-04-22',
            '2011-05-18',
            '2014-05-07',
            '2016-08-03',
            '2019-05-08',
            # In 2008, Human Rights Day fell on the same day as Good
            # Friday (March 21), so the "Bridge Public Holiday" was
            # observed on May 2.
            '2008-05-02',
            # In 2011 and 2016, Christmas fell on Sunday, which would
            # ordinarily cause it to be observed on Monday; but Day of
            # Goodwill fell on Monday, which resulted in only 11
            # scheduled holidays for the year instead of the usual 12.
            # In these years, the 27th was declared a bonus holiday: it
            # may or may not be repeated in 2022 and onward.
            '2011-12-27',
            '2016-12-27',
        ]]



# -*- coding: utf-8 -*-
#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta
from itertools import chain

import numpy as np
import pandas as pd
from pandas.tseries.holiday import (
    EasterMonday,
    GoodFriday,
    Holiday,
    sunday_to_monday,
)
from pandas.tseries.offsets import (
    LastWeekOfMonth,
    WeekOfMonth,
)
from pytz import timezone
import toolz

from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
    MONDAY,
    TUESDAY,
    WEDNESDAY,
    THURSDAY,
    FRIDAY,
    SATURDAY,
    SUNDAY,
)
from .common_holidays import (
    boxing_day,
    christmas,
    christmas_eve,
    new_years_day,
    new_years_eve,
    weekend_christmas,
)
from .lunisolar_holidays import (
    chinese_buddhas_birthday_dates,
    chinese_lunar_new_year_dates,
    mid_autumn_festival_dates,
    double_ninth_festival_dates,
    dragon_boat_festival_dates,
    qingming_festival_dates
)
from .utils.pandas_utils import vectorized_sunday_to_monday

# Useful resources for making changes to this file:
# # /etc/lunisolar
# http://www.math.nus.edu.sg/aslaksen/calendar/cal.pdf
# https://www.hko.gov.hk/gts/time/calendarinfo.htm
#   - the almanacs on this page are also useful

weekdays = (MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY)


def process_queen_birthday(dt):
    # before 1983
    if dt.year in [1974, 1981]:
        return dt + pd.DateOffset(weekday=6)
    elif dt.year < 1983:
        return sunday_to_monday(dt)
    # after 1983
    wom = WeekOfMonth(week=2, weekday=0)
    if dt.year in [1983, 1988, 1993, 1994]:
        wom = WeekOfMonth(week=1, weekday=0)
    if dt.year in [1985]:
        wom = WeekOfMonth(week=3, weekday=0)
    return dt + wom


LabourDay = Holiday(
    name="Labour Day",  # 劳动节
    month=5,
    day=1,
    observance=sunday_to_monday,
    start_date=pd.Timestamp('1999-05-01')
)

HKRegionEstablishmentDay = Holiday(
    name="Hong Kong Special Region Establishment Day",
    month=7,
    day=1,
    observance=sunday_to_monday,
    start_date=pd.Timestamp('1997-07-01')
)

NationalDay = Holiday(
    name="National Day",
    month=10,
    day=1,
    observance=sunday_to_monday,
    start_date=pd.Timestamp('1997-07-01')
)

QueenBirthday = Holiday(
    name="Queen's Birthday",  # 英女王生日 6月
    month=6,
    day=10,
    observance=process_queen_birthday,
    start_date=pd.Timestamp('1983-01-01'),
    end_date=pd.Timestamp('1997-06-01')
)

QueenBirthday2 = Holiday(
    name="Queen's Birthday",  # 英女王生日 4月
    month=4,
    day=21,
    observance=process_queen_birthday,
    start_date=pd.Timestamp('1926-04-21'),
    end_date=pd.Timestamp('1983-01-01')
)

CommemoratingAlliedVictory = Holiday(
    name="Commemorating the allied victory",  # 重光纪念日 8月最后一个星期一
    month=8,
    day=20,
    offset=LastWeekOfMonth(weekday=0),
    start_date=pd.Timestamp('1945-08-30'),
    end_date=pd.Timestamp('1997-07-01')
)

IDontKnow = Holiday(
    name="I dont know these days, please tell me",  # 8月第一个星期一
    month=7,
    day=31,
    offset=WeekOfMonth(week=0, weekday=0),
    start_date=pd.Timestamp('1960-08-01'),
    end_date=pd.Timestamp('1983-01-01')
)


day_after_mid_autumn_festival_dates = mid_autumn_festival_dates + timedelta(1)

HKAdhocClosures = [
    # I dont know these days
    pd.Timestamp('1970-07-01', tz='UTC'),
    pd.Timestamp('1971-07-01', tz='UTC'),
    pd.Timestamp('1973-07-02', tz='UTC'),
    pd.Timestamp('1974-07-01', tz='UTC'),
    pd.Timestamp('1975-07-01', tz='UTC'),
    pd.Timestamp('1976-07-01', tz='UTC'),
    pd.Timestamp('1977-07-01', tz='UTC'),
    pd.Timestamp('1979-07-02', tz='UTC'),
    pd.Timestamp('1980-07-01', tz='UTC'),
    pd.Timestamp('1981-07-01', tz='UTC'),
    pd.Timestamp('1982-07-01', tz='UTC'),
    pd.Timestamp('1971-03-22', tz='UTC'),
    pd.Timestamp('1971-12-06', tz='UTC'),
    pd.Timestamp('1971-12-20', tz='UTC'),
    pd.Timestamp('1975-07-28', tz='UTC'),
    pd.Timestamp('1985-07-29', tz='UTC'),

    # Weather related closures
    pd.Timestamp('1970-07-16', tz='UTC'),  # 台风Ruby7003
    pd.Timestamp('1970-09-14', tz='UTC'),  # 台风Georgia7011
    pd.Timestamp('1971-07-22', tz='UTC'),  # 台风Lucy7114
    pd.Timestamp('1971-08-31', tz='UTC'),  # 重光纪念日?
    pd.Timestamp('1973-04-16', tz='UTC'),  # 股灾休市?
    pd.Timestamp('1973-07-17', tz='UTC'),  # 台风Dot7304
    pd.Timestamp('1974-04-25', tz='UTC'),  # 英国女王生日
    pd.Timestamp('1975-10-14', tz='UTC'),  # 台风Elsie7514
    pd.Timestamp('1978-07-26', tz='UTC'),  # 台风Agnes7807
    pd.Timestamp('1978-07-27', tz='UTC'),
    pd.Timestamp('1979-01-26', tz='UTC'),  # 春节补假
    pd.Timestamp('1979-08-02', tz='UTC'),  # 台风Hope7908
    pd.Timestamp('1980-05-21', tz='UTC'),  # 台风Georgia8004
    pd.Timestamp('1980-07-22', tz='UTC'),  # 台风Joy8007
    pd.Timestamp('1981-04-27', tz='UTC'),  # 英国女王生日
    pd.Timestamp('1981-07-06', tz='UTC'),  # 台风Lynn8106
    pd.Timestamp('1981-07-07', tz='UTC'),
    pd.Timestamp('1981-07-29', tz='UTC'),  # 查理斯王子与戴安娜婚礼
    pd.Timestamp('1983-09-09', tz='UTC'),  # 台风Ellen8309
    pd.Timestamp('1985-06-24', tz='UTC'),  # 台风Hal8504
    pd.Timestamp('1986-04-01', tz='UTC'),  # 复活节星期一翌日
    pd.Timestamp('1986-10-22', tz='UTC'),  # 英女王伊丽莎白二世访港
    pd.Timestamp('1987-10-20', tz='UTC'),  # 黑色星期一后,休市4天
    pd.Timestamp('1987-10-21', tz='UTC'),
    pd.Timestamp('1987-10-22', tz='UTC'),
    pd.Timestamp('1987-10-23', tz='UTC'),
    pd.Timestamp('1988-04-05', tz='UTC'),  # 清明节翌日
    # Timestamp('1988-06-13', tz='UTC'),  # 英国女王生日
    pd.Timestamp('1991-06-18', tz='UTC'),  # 英国女王生日翌日
    pd.Timestamp('1992-07-22', tz='UTC'),  # 台风Cary9207
    # pd.Timestamp('1993-06-14', tz='UTC'),  # 英国女王生日
    pd.Timestamp('1993-09-17', tz='UTC'),  # 台风Becky9316
    pd.Timestamp('1994-06-14', tz='UTC'),  # 英国女王生日翌日,端午节翌日
    pd.Timestamp('1997-06-30', tz='UTC'),  # 英国女王生日
    pd.Timestamp('1997-07-02', tz='UTC'),  # 香港回归纪念日翌日
    pd.Timestamp('1997-08-18', tz='UTC'),  # 抗战胜利纪念日
    pd.Timestamp('1997-10-02', tz='UTC'),  # 国庆节翌日
    pd.Timestamp('1998-08-17', tz='UTC'),  # 抗战胜利纪念日
    pd.Timestamp('1998-10-02', tz='UTC'),  # 国庆节翌日
    pd.Timestamp('1999-04-06', tz='UTC'),  # 清明节翌日
    pd.Timestamp('1999-09-16', tz='UTC'),  # 台风约克
    pd.Timestamp('1999-12-31', tz='UTC'),  # 千年虫
    pd.Timestamp('2001-07-06', tz='UTC'),  # 台风尤特0104
    pd.Timestamp('2001-07-25', tz='UTC'),  # 台风玉兔0107
    # pd.Timestamp(2008-06-25', tz='UTC'),  # 台风风神0806,上午休市
    pd.Timestamp('2008-08-06', tz='UTC'),  # 台风北冕0809
    pd.Timestamp('2008-08-22', tz='UTC'),  # 台风鹦鹉0810
    # pd.Timestamp(2009-09-15', tz='UTC'),  # 台风巨爵0915,上午休市
    pd.Timestamp('2010-04-06', tz='UTC'),  # 清明节翌日
    pd.Timestamp('2011-09-29', tz='UTC'),  # 台风纳沙1117
    # pd.Timestamp(2012-07-24', tz='UTC'),  # 台风韦森特1208,上午休市
    pd.Timestamp('2012-10-02', tz='UTC'),  # 中秋节补假
    # pd.Timestamp(2013-05-22', tz='UTC'),  # 暴雨,上午休市
    pd.Timestamp('2013-08-14', tz='UTC'),  # 台风尤特1311
    # pd.Timestamp(2013-09-23', tz='UTC'),  # 台风天兔1319,上午休市
    # pd.Timestamp(2014-09-16', tz='UTC'),  # 台风海鸥1415,上午休市
    pd.Timestamp('2015-04-07', tz='UTC'),  # 复活节+清明节补假
    # pd.Timestamp(2015-07-09', tz='UTC'),  # 台风莲花1520,期货夜盘休市
    pd.Timestamp('2015-09-03', tz='UTC'),  # 抗战70周年纪念
    # pd.Timestamp(2016-08-01', tz='UTC'),  # 台风妮妲1604,期货夜盘20:55收市
    pd.Timestamp('2016-08-02', tz='UTC'),  # 台风妮妲1604
    pd.Timestamp('2016-10-21', tz='UTC'),  # 台风海马1622
    # pd.Timestamp(2017-06-12', tz='UTC'),  # 台风苗柏1702,期货夜盘17:35休市
    pd.Timestamp('2017-08-23', tz='UTC'),  # 台风天鸽1713
]


def boxing_day_obs(dt):
    if dt.weekday in (MONDAY, TUESDAY):
        return dt + pd.Timedelta(days=1)
    return dt


class XHKGExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Hong Kong Stock Exchange (XHKG).

    Open Time: 9:31 AM, Asia/Hong_Kong
    Lunch Break: 12:01 PM - 1:00 PM Asia/Hong_Kong
    Close Time: 4:00 PM, Asia/Hong_Kong

    Regularly-Observed Holidays:
    - New Years Day (observed on monday when Jan 1 is a Sunday)
    - Lunar New Year and the following 2 days. If the 3rd day of the lunar year
      is a Sunday, then the next Monday is a holiday.
    - Ching Ming Festival
    - Good Friday
    - Easter Monday
    - Buddhas Birthday
    - Dragon Boat Festival
    - Chinese National Day (observed on monday when Oct 1 is a Sunday)
    - Day Following Mid-Autumn Festival
    - Chung Yeung Festival
    - Christmas (observed on nearest weekday to December 25)
    - Day after Christmas is observed

    Regularly-Observed Early Closes:
    - Lunar New Year's Eve
    - Christmas Eve
    - New Year's Eve


    Additional Irregularities:
    - Closes frequently for severe weather.
    """
    name = 'XHKG'
    tz = timezone('Asia/Hong_Kong')

    open_times = (
        (None, time(10, 1)),
        (pd.Timestamp('2011-03-07'), time(9, 31)),
    )
    break_start_times = (
        (None, time(12, 1)),
    )
    break_end_times = (
        (None, time(13, 0)),
    )
    close_times = (
        (None, time(16)),
    )
    regular_early_close_times = (
        (None, time(12, 30)),
        (pd.Timestamp('2011-03-07'), time(12, 00)),
    )

    def __init__(self, *args, **kwargs):
        super(XHKGExchangeCalendar, self).__init__(*args, **kwargs)

        lunisolar_holidays = (
            chinese_buddhas_birthday_dates,
            chinese_lunar_new_year_dates,
            day_after_mid_autumn_festival_dates,
            double_ninth_festival_dates,
            dragon_boat_festival_dates,
            qingming_festival_dates,
        )
        earliest_precomputed_year = max(map(np.min, lunisolar_holidays)).year
        if earliest_precomputed_year > self.first_trading_session.year:
            raise ValueError(
                'the lunisolar holidays have only been computed back to {},'
                ' cannot instantiate the XHKG calendar back to {}'.format(
                    earliest_precomputed_year,
                    self.first_trading_session.year,
                ),
            )

        latest_precomputed_year = min(map(np.max, lunisolar_holidays)).year
        if latest_precomputed_year < self.last_trading_session.year:
            raise ValueError(
                'the lunisolar holidays have only been computed through {},'
                ' cannot instantiate the XHKG calendar in {}'.format(
                    latest_precomputed_year,
                    self.last_trading_session.year,
                ),
            )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            new_years_day(observance=sunday_to_monday),
            GoodFriday,
            EasterMonday,
            LabourDay,
            HKRegionEstablishmentDay,
            CommemoratingAlliedVictory,
            IDontKnow,
            NationalDay,
            QueenBirthday,
            QueenBirthday2,
            christmas(),
            weekend_christmas(),
            boxing_day(observance=boxing_day_obs)
        ])

    @property
    def adhoc_holidays(self):
        lunar_new_years_eve = (
            chinese_lunar_new_year_dates - pd.Timedelta(days=1)
        )[
            (chinese_lunar_new_year_dates.weekday == SATURDAY) &
            (chinese_lunar_new_year_dates.year < 2013)
        ]

        lunar_new_year_2 = chinese_lunar_new_year_dates + pd.Timedelta(days=1)
        lunar_new_year_3 = chinese_lunar_new_year_dates + pd.Timedelta(days=2)
        lunar_new_year_4 = (
            chinese_lunar_new_year_dates + pd.Timedelta(days=3)
        )[
            # According to the new arrangement under the General Holidays and
            # Employment Legislation (Substitution of Holidays)(Amendment)
            # Ordinance 2011, when either Lunar New Year's Day, the second day
            # of Lunar New Year or the third day of Lunar New Year falls on a
            # Sunday, the fourth day of Lunar New Year is designated as a
            # statutory and general holiday in substitution.
            (
                (chinese_lunar_new_year_dates.weekday == SUNDAY) |
                (lunar_new_year_2.weekday == SUNDAY) |
                (lunar_new_year_3.weekday == SUNDAY)
            ) &
            (chinese_lunar_new_year_dates.year >= 2013)
        ]

        qingming_festival = vectorized_sunday_to_monday(
            qingming_festival_dates,
        ).values
        years = qingming_festival.astype('M8[Y]')
        easter_monday = EasterMonday.dates(years[0], years[-1] + 1)
        # qingming gets observed one day later if easter monday is on the same
        # day
        qingming_festival[qingming_festival == easter_monday] += (
            np.timedelta64(1, 'D')
        )

        # if the day after the mid autumn festival is October first, which
        # conflicts with national day, then national day is observed on the
        # second, though we don't encode that in the regular holiday, so
        # instead we pretend that the mid autumn festival would be delayed.
        mid_autumn_festival = day_after_mid_autumn_festival_dates.values
        mid_autumn_festival[
            (day_after_mid_autumn_festival_dates.month == 10) &
            (day_after_mid_autumn_festival_dates.day == 1)
        ] += np.timedelta64(1, 'D')

        return list(chain(
            lunar_new_years_eve,
            chinese_lunar_new_year_dates,
            lunar_new_year_2,
            lunar_new_year_3,
            lunar_new_year_4,

            qingming_festival,
            vectorized_sunday_to_monday(chinese_buddhas_birthday_dates),
            vectorized_sunday_to_monday(dragon_boat_festival_dates),
            mid_autumn_festival,
            vectorized_sunday_to_monday(double_ninth_festival_dates),
            HKAdhocClosures,
        ))

    @property
    def special_closes(self):
        return [
            # HK changed their early close time
            (
                time(12, 30), HolidayCalendar([
                    new_years_eve(
                        # Market was close for Y2K instead of closing early
                        end_date=pd.Timestamp('1999-12-01'),
                        days_of_week=weekdays,
                    ),
                    new_years_eve(
                        start_date=pd.Timestamp('2000-12-01'),
                        end_date=pd.Timestamp('2011-03-07'),
                        days_of_week=weekdays,
                    ),
                    christmas_eve(
                        end_date=pd.Timestamp('2011-03-07'),
                        days_of_week=weekdays
                    )])
            ),
            (
                time(12, 00), HolidayCalendar([
                    new_years_eve(
                        start_date=pd.Timestamp('2011-03-07'),
                        days_of_week=weekdays,
                    ),
                    christmas_eve(
                        start_date=pd.Timestamp('2011-03-07'),
                        days_of_week=weekdays
                    )])
            ),
        ]

    @property
    def special_closes_adhoc(self):
        lunar_new_years_eve = (
            chinese_lunar_new_year_dates - pd.Timedelta(days=1)
        )[
            np.in1d(
                chinese_lunar_new_year_dates.weekday,
                [TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY],
            ) & (chinese_lunar_new_year_dates.year >= 2013)
        ].values

        def selection(arr, start, end):
            predicates = []
            if start is not None:
                predicates.append(start.asm8 <= arr)
            if end is not None:
                predicates.append(arr < end.asm8)

            if not predicates:
                return arr

            return arr[np.all(predicates, axis=0)]

        return [
            (time, selection(lunar_new_years_eve, start, end))
            for (start, time), (end, _) in toolz.sliding_window(
                2,
                toolz.concatv(self.regular_early_close_times, [(None, None)]),
            )
        ]



from datetime import timedelta

import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    next_monday_or_tuesday,
    sunday_to_monday,
    weekend_to_monday,
)
from pytz import UTC

from .common_holidays import european_labour_day, new_years_day, new_years_eve
from .trading_calendar import SUNDAY, MONDAY


def new_years_eve_observance(holidays):
    # New Year's Eve is a holiday every year except for 2003 for some reason.
    holidays = holidays[holidays.year != 2003]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


def new_years_day_observance(holidays):
    # There was no extra observance of New Year's Day in 2006.
    holidays = holidays[holidays.year != 2006]

    return pd.to_datetime([next_monday_or_tuesday(day) for day in holidays])


def songkran_festival_last_day_observance(dt):
    """
    This function is similar to the pandas function `next_monday_or_tuesday`
    except it does not observe Saturday holidays on Monday.
    """
    if dt.weekday() == SUNDAY or dt.weekday() == MONDAY:
        return dt + timedelta(days=1)
    return dt


NewYearsDay = new_years_day(observance=new_years_day_observance)

ChakriMemorialDay = Holiday(
    'Chakri Memorial Day',
    month=4,
    day=6,
    observance=weekend_to_monday,
)

# Thai New Year. This does not follow the usual observe-next-trading-day rule.
SongkranFestival1 = Holiday('Songkran Festival', month=4, day=13)
SongkranFestival2 = Holiday(
    'Songkran Festival',
    month=4,
    day=14,
    observance=sunday_to_monday,
)
SongkranFestival3 = Holiday(
    'Songkran Festival',
    month=4,
    day=15,
    observance=songkran_festival_last_day_observance,
)

LabourDay = european_labour_day(observance=weekend_to_monday)

CoronationDay2016AndBefore = Holiday(
    'Coronation Day For King #9',
    month=5,
    day=5,
    observance=weekend_to_monday,
    end_date='2017',
)
CoronationDay2019AndAfter = Holiday(
    'Coronation Day For King #10',
    month=5,
    day=4,
    observance=weekend_to_monday,
    start_date='2019',
)

HMQueensBirthday = Holiday(
    "Her Majesty The Queen's Birthday",
    month=6,
    day=3,
    observance=weekend_to_monday,
    start_date='2019',
)
HMKingsBirthday = Holiday(
    "His Majesty The King's Birthday",
    month=7,
    day=28,
    observance=weekend_to_monday,
    start_date='2017',
)
HMQueenMothersBirthday = Holiday(
    "Her Majesty The Queen Mother's Birthday",
    month=8,
    day=12,
    observance=weekend_to_monday,
)

# This holiday was historically used as a "catch up" day for the exchange, so
# it does not need to follow the usual observe-next-trading-day rule.
HalfYearHoliday = Holiday(
    'Half Year Holiday',
    month=7,
    day=1,
    start_date='2002',
    end_date='2017',
)

ThePassingOfKingBhumibol = Holiday(
    'The Passing of King Bhumibol',
    month=10,
    day=13,
    observance=weekend_to_monday,
    start_date='2017',
)

ChulalongkornDay = Holiday(
    'Chulalongkorn Day',
    month=10,
    day=23,
    observance=weekend_to_monday,
)

KingBhumibolsBirthday = Holiday(
    "King Bhumibol's Birthday",
    month=12,
    day=5,
    observance=weekend_to_monday,
)

ThailandConstitutionDay = Holiday(
    'Thailand Constitution Day',
    month=12,
    day=10,
    observance=weekend_to_monday,
)

NewYearsEve = new_years_eve(observance=new_years_eve_observance)

# Adhoc Holidays
# --------------

new_years_bridge_days = [
    pd.Timestamp('2002-12-30', tz=UTC),
    pd.Timestamp('2004-01-02', tz=UTC),
    pd.Timestamp('2009-01-02', tz=UTC),
    pd.Timestamp('2013-12-30', tz=UTC),
    pd.Timestamp('2015-01-02', tz=UTC),
]

asanha_bucha_bridge_days = [
    pd.Timestamp('2009-07-06', tz=UTC),
    pd.Timestamp('2016-07-18', tz=UTC),
]

queens_birthday_bridge_days = [
    pd.Timestamp('2010-08-13', tz=UTC),
    pd.Timestamp('2014-08-11', tz=UTC),
]

coronation_bridge_days = [
    pd.Timestamp('2015-05-04', tz=UTC),
    pd.Timestamp('2016-05-06', tz=UTC),
]

vesak_bridge_days = [
    pd.Timestamp('2011-05-16', tz=UTC),
]

misc_adhoc = [
    pd.Timestamp('2006-04-19', tz=UTC),  # Special Holiday
    pd.Timestamp('2006-06-12', tz=UTC),  # Special Holiday
    pd.Timestamp('2006-06-13', tz=UTC),  # Special Holiday
    pd.Timestamp('2006-09-20', tz=UTC),  # Exchange Holiday
    pd.Timestamp('2007-12-24', tz=UTC),  # Exchange Holiday
    pd.Timestamp('2010-05-20', tz=UTC),  # Closure Due to Security Concerns
    pd.Timestamp('2010-05-21', tz=UTC),  # Closure Due to Security Concerns
    pd.Timestamp('2012-04-09', tz=UTC),  # Bank Holiday
    pd.Timestamp('2017-10-26', tz=UTC),  # Cremation of King Bhumibol
]

# Lunar Holidays
# --------------

# Makha Bucha (also known as Magha Puja) is celebrated on the day of the Full
# Moon of Magha in the Buddhist calendar. This falls sometime between February
# and March.
makha_bucha = pd.to_datetime([
    '1981-02-18',
    '1982-02-08',
    '1983-02-27',
    '1984-02-16',
    '1985-03-06',
    '1986-02-24',
    '1987-02-13',
    '1988-03-03',
    '1989-02-20',
    '1990-02-09',
    '1991-02-28',
    '1992-02-18',
    '1993-03-08',
    '1994-02-25',
    '1995-02-15',
    '1996-03-05',
    '1997-02-22',
    '1998-02-11',
    '1999-03-02',
    '2000-02-19',
    '2001-02-08',
    '2002-02-26',
    '2003-02-17',
    '2004-03-05',
    '2005-02-23',
    '2006-02-13',
    '2007-03-05',
    '2008-02-21',
    '2009-02-09',
    '2010-03-01',
    '2011-02-18',
    '2012-03-07',
    '2013-02-25',
    '2014-02-14',
    '2015-03-04',
    '2016-02-22',
    '2017-02-13',
    '2018-03-01',
    '2019-02-19',
    '2020-02-10',
])

# Vesak (also known as Buddha Day) is celebrated on the day of the Full Moon of
# Visakha in the Buddhist calendar. This typically falls in May.
vesak = pd.to_datetime([
    '1981-05-18',
    '1982-05-07',
    '1983-05-26',
    '1984-05-15',
    '1985-06-02',
    '1986-05-23',
    '1987-05-13',
    '1988-05-31',
    '1989-05-20',
    '1990-05-09',
    '1991-05-28',
    '1992-05-16',
    '1993-06-04',
    '1994-05-24',
    '1995-05-14',
    '1996-06-01',
    '1997-05-22',
    '1998-05-11',
    '1999-05-30',
    '2000-05-18',
    '2001-05-07',
    '2002-05-27',
    '2003-05-15',
    '2004-06-02',
    '2005-05-23',
    '2006-05-12',
    '2007-05-31',
    '2008-05-19',
    '2009-05-08',
    '2010-05-28',
    '2011-05-17',
    '2012-06-04',
    '2013-05-24',
    '2014-05-13',
    '2015-06-01',
    '2016-05-20',
    '2017-05-10',
    '2018-05-29',
    '2019-05-20',
    '2020-05-06',
])

# Asanha Bucha (also known as Asalha Puja) is celebrated on the day of the Full
# Moon of Asadha in the Buddhist calendar. This typically falls in July.
asanha_bucha = pd.to_datetime([
    '1981-07-17',
    '1982-07-06',
    '1983-07-24',
    '1984-07-12',
    '1985-07-31',
    '1986-07-21',
    '1987-07-10',
    '1988-07-28',
    '1989-07-18',
    '1990-07-07',
    '1991-07-26',
    '1992-07-14',
    '1993-08-02',
    '1994-07-22',
    '1995-07-12',
    '1996-07-30',
    '1997-07-19',
    '1998-07-09',
    '1999-07-28',
    '2000-07-16',
    '2001-07-05',
    '2002-07-25',
    '2003-07-14',
    '2004-08-02',
    '2005-07-22',
    '2006-07-11',
    '2007-07-30',
    '2008-07-17',
    '2009-07-07',
    '2010-07-26',
    '2011-07-15',
    '2012-08-02',
    '2013-07-22',
    '2014-07-11',
    '2015-07-30',
    '2016-07-19',
    '2017-07-10',
    '2018-07-27',
    '2019-07-16',
    '2020-07-06',
])



import pandas as pd
from pandas.tseries.holiday import DateOffset, Easter, FR, Holiday
from pandas.tseries.offsets import Day

from .trading_calendar import MONDAY, TUESDAY


def new_years_day(start_date=None,
                  end_date=None,
                  observance=None,
                  days_of_week=None):
    return Holiday(
        "New Year's Day",
        month=1,
        day=1,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def new_years_eve(start_date=None,
                  end_date=None,
                  observance=None,
                  days_of_week=None):
    return Holiday(
        "New Year's Eve",
        month=12,
        day=31,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def epiphany(start_date=None,
             end_date=None,
             observance=None,
             days_of_week=None):
    return Holiday(
        'Epiphany',
        month=1,
        day=6,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def anzac_day(start_date=None,
              end_date=None,
              observance=None,
              days_of_week=None):
    return Holiday(
        'Anzac Day',
        month=4,
        day=25,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def european_labour_day(start_date=None,
                        end_date=None,
                        observance=None,
                        days_of_week=None):
    return Holiday(
        "Labour Day",
        month=5,
        day=1,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


# Holy Wednesday, Maundy Thursday, Ascension Day, Whit Monday, and Corpus
# Christi do not take observance as a parameter because they depend on a
# particular offset, and offset and observance cannot both be passed to a
# Holiday.
def holy_wednesday(start_date=None, end_date=None, days_of_week=None):
    return Holiday(
        'Holy Wednesday',
        month=1,
        day=1,
        offset=[Easter(), -Day(4)],
        start_date=start_date,
        end_date=end_date,
        days_of_week=days_of_week,
    )


def maundy_thursday(start_date=None, end_date=None, days_of_week=None):
    return Holiday(
        'Maundy Thursday',
        month=1,
        day=1,
        offset=[Easter(), -Day(3)],
        start_date=start_date,
        end_date=end_date,
        days_of_week=days_of_week,
    )


def ascension_day(start_date=None, end_date=None):
    return Holiday(
        "Ascension Day",
        month=1,
        day=1,
        offset=[Easter(), Day(39)],
        start_date=start_date,
        end_date=end_date,
    )


def whit_monday(start_date=None, end_date=None):
    return Holiday(
        "Whit Monday",
        month=1,
        day=1,
        offset=[Easter(), Day(50)],
        start_date=start_date,
        end_date=end_date,
    )


def corpus_christi(start_date=None, end_date=None):
    return Holiday(
        'Corpus Christi',
        month=1,
        day=1,
        offset=[Easter(), Day(60)],
        start_date=start_date,
        end_date=end_date,
    )


def midsummer_eve(start_date=None, end_date=None):
    return Holiday(
        'Midsummer Eve',
        month=6,
        day=19,
        offset=DateOffset(weekday=FR(1)),
        start_date=start_date,
        end_date=end_date,
    )


def saint_peter_and_saint_paul_day(start_date=None,
                                   end_date=None,
                                   observance=None,
                                   days_of_week=None):
    return Holiday(
        'Saint Peter and Saint Paul Day',
        month=6,
        day=29,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def assumption_day(start_date=None,
                   end_date=None,
                   observance=None,
                   days_of_week=None):
    return Holiday(
        'Assumption Day',
        month=8,
        day=15,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def all_saints_day(start_date=None,
                   end_date=None,
                   observance=None,
                   days_of_week=None):
    return Holiday(
        'All Saints Day',
        month=11,
        day=1,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def immaculate_conception(start_date=None,
                          end_date=None,
                          observance=None,
                          days_of_week=None):
    return Holiday(
        'Immaculate Conception',
        month=12,
        day=8,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def christmas_eve(start_date=None,
                  end_date=None,
                  observance=None,
                  days_of_week=None):
    return Holiday(
        'Christmas Eve',
        month=12,
        day=24,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def christmas(start_date=None,
              end_date=None,
              observance=None,
              days_of_week=None):
    return Holiday(
        "Christmas",
        month=12,
        day=25,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def weekend_christmas(start_date=None, end_date=None, observance=None):
    """
    If christmas day is Saturday Monday 27th is a holiday
    If christmas day is sunday the Tuesday 27th is a holiday
    """
    return Holiday(
        "Weekend Christmas",
        month=12,
        day=27,
        days_of_week=(MONDAY, TUESDAY),
        start_date=start_date,
        end_date=end_date,
        observance=observance,
    )


def boxing_day(start_date=None,
               end_date=None,
               observance=None,
               days_of_week=None):
    return Holiday(
        "Boxing Day",
        month=12,
        day=26,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def weekend_boxing_day(start_date=None, end_date=None, observance=None):
    """
    If boxing day is saturday then Monday 28th is a holiday
    If boxing day is sunday then Tuesday 28th is a holiday
    """
    return Holiday(
        "Weekend Boxing Day",
        month=12,
        day=28,
        days_of_week=(MONDAY, TUESDAY),
        start_date=start_date,
        end_date=end_date,
        observance=observance,
    )


# These dates were initially calculated using the ummalqura Python
# package (https://pypi.org/project/ummalqura/), and then tweaked
# to fit Turkey's observance of Eid al-Fitr.  Other countries that
# observe Eid al-Fitr might use slightly different dates
eid_al_fitr_first_day = pd.to_datetime([
    '1981-08-01',
    '1982-07-21',
    '1983-07-11',
    '1984-06-30',
    '1985-06-19',
    '1986-06-08',
    '1987-05-28',
    '1988-05-16',
    '1989-05-06',
    '1990-04-26',
    '1991-04-15',
    '1992-04-04',
    '1993-03-24',
    '1994-03-13',
    '1995-03-02',
    '1996-02-19',
    '1997-02-08',
    '1998-01-29',
    '1999-01-18',
    '2000-01-08',
    '2000-12-27',
    '2001-12-16',
    '2002-12-05',
    '2003-11-25',
    '2004-11-14',
    '2005-11-03',
    '2006-10-23',
    '2007-10-12',
    '2008-09-30',
    '2009-09-20',
    '2010-09-09',
    '2011-08-30',
    '2012-08-19',
    '2013-08-08',
    '2014-07-28',
    '2015-07-17',
    '2016-07-05',
    '2017-06-25',
    '2018-06-15',
    '2019-06-04',
    '2020-05-24',
    '2021-05-13',
    '2022-05-02',
    '2023-04-21',
    '2024-04-10',
    '2025-03-30',
    '2026-03-20',
    '2027-03-09',
    '2028-02-26',
    '2029-02-14',
    '2030-02-04',
    '2031-01-24',
    '2032-01-14',
    '2033-01-02',
    '2033-12-23',
    '2034-12-12',
    '2035-12-01',
    '2036-11-19',
    '2037-11-08',
    '2038-10-29',
    '2039-10-19',
    '2040-10-07',
    '2041-09-26',
    '2042-09-15',
    '2043-09-04',
    '2044-08-24',
    '2045-08-14',
    '2046-08-03',
    '2047-07-24',
    '2048-07-12',
    '2049-07-01',
])

# These dates were initially calculated using the ummalqura Python
# package (https://pypi.org/project/ummalqura/), and then tweaked
# to fit Turkey's observance of Eid al-Adha.  Other countries that
# observe Eid al-Adha might use slightly different dates
eid_al_adha_first_day = pd.to_datetime([
    '1981-10-08',
    '1982-09-27',
    '1983-09-17',
    '1984-09-05',
    '1985-08-26',
    '1986-08-15',
    '1987-08-04',
    '1988-07-23',
    '1989-07-13',
    '1990-07-02',
    '1991-06-22',
    '1992-06-11',
    '1993-05-31',
    '1994-05-20',
    '1995-05-09',
    '1996-04-27',
    '1997-04-17',
    '1998-04-07',
    '1999-03-27',
    '2000-03-16',
    '2001-03-05',
    '2002-02-22',
    '2003-02-10',
    '2004-02-01',
    '2005-01-20',
    '2006-01-09',
    '2006-12-31',
    '2007-12-20',
    '2008-12-08',
    '2009-11-27',
    '2010-11-16',
    '2011-11-06',
    '2012-10-25',
    '2013-10-15',
    '2014-10-04',
    '2015-09-24',
    '2016-09-12',
    '2017-09-01',
    '2018-08-21',
    '2019-08-11',
    '2020-07-31',
    '2021-07-20',
    '2022-07-09',
    '2023-06-28',
    '2024-06-16',
    '2025-06-06',
    '2026-05-27',
    '2027-05-16',
    '2028-05-05',
    '2029-04-24',
    '2030-04-13',
    '2031-04-02',
    '2032-03-22',
    '2033-03-11',
    '2034-03-01',
    '2035-02-18',
    '2036-02-07',
    '2037-01-26',
    '2038-01-16',
    '2039-01-05',
    '2039-12-26',
    '2040-12-14',
    '2041-12-04',
    '2042-11-23',
    '2043-11-12',
    '2044-10-31',
    '2045-10-21',
    '2046-10-10',
    '2047-09-30',
    '2048-09-19',
    '2049-09-08',
])



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain
import pandas as pd
from pytz import timezone

from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
)
from .xkls_holidays import (
    NewYearsDay,
    Thaipusam,
    FederalTerritoryDay,
    ChineseNewYear,
    ChineseNewYearDay2,
    ChineseNewYearsHalfDay,
    LabourDay,
    WesakDay,
    NuzulAlQuran,
    EidAlFitrDay1,
    EidAlFitrDay2,
    EidAlFitrHalfDay,
    EidAlAdha,
    NationalDay,
    Muharram,
    MalaysiaDay,
    Deepavali,
    MuhammadBirthday,
    ChristmasDay,
    misc_adhoc,
)


class XKLSExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Malaysia Stock Exchange (XKLS).

    Open Time: 9:00 AM, MYT
    Close Time: 5:00 PM, MYT

    Regularly-Observed Holidays:
    - New Year's Day
    - Thaipusam (Tamil Calendar)
    - Federal Territory Day (Feb 1)
    - Chinese New Year (2 days)
    - Labour Day (May 1)
    - Wesak Day (Chinese Lunar Calendar)
    - Nuzul Al'Quran (Islamic Lunar Calendar)
    - Eid al-Fitr (2 days)
    - Eid al-Adha
    - National Day (Aug 31)
    - Muharram (Lunar Calendar)
    - Malaysia Day (Sep 16)
    - Deepavali
    - Prophet Muhammad's Birthday
    - Christmas Day

    Early Closes:
    - Chinese New Year's Eve
    - Eid al-Fitr Eve

    TODO: XKLS takes a two hour lunch break from 12:30-2:30pm, but we are
    ignoring this for now.
    """
    name = 'XKLS'

    tz = timezone('Asia/Kuala_Lumpur')

    open_times = (
        (None, time(9, 1)),
    )

    close_times = (
        (None, time(17, 00)),
    )

    regular_early_close = time(12, 30)

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            FederalTerritoryDay,
            LabourDay,
            NationalDay,
            MalaysiaDay,
            ChristmasDay,
        ])

    @property
    def adhoc_holidays(self):

        return list(chain(
            misc_adhoc,
            ChineseNewYear,
            ChineseNewYearDay2,
            NuzulAlQuran,
            EidAlFitrDay1,
            EidAlFitrDay2,
            EidAlAdha,
            Muharram,
            MuhammadBirthday,
            Deepavali,
            Thaipusam,
            WesakDay,
        ))

    @property
    def special_closes_adhoc(self):
        # Regular early closes on Chinese New Years Eve, Eid al-Fitr Eve
        early_close_days = ChineseNewYearsHalfDay + EidAlFitrHalfDay

        collisions = [
            # Chinese New Year's Eve was a holiday until 2005
            pd.Timestamp('1997-02-07'),
            pd.Timestamp('1998-01-28'),
            pd.Timestamp('2002-02-11'),
            pd.Timestamp('2003-01-31'),
            pd.Timestamp('2004-01-21'),
            # Eid al-Fitr Eve was a holiday in 2003
            pd.Timestamp('2003-11-24'),
        ]

        early_close_days = list(set(early_close_days) - set(collisions))

        return [
            (self.regular_early_close, early_close_days)
        ]



import numpy as np
import pandas as pd

NANOSECONDS_PER_MINUTE = int(6e10)

NP_NAT = np.array([pd.NaT], dtype=np.int64)[0]


def next_divider_idx(dividers, minute_val):

    divider_idx = np.searchsorted(dividers, minute_val, side="right")
    target = dividers[divider_idx]

    if minute_val == target:
        # if dt is exactly on the divider, go to the next value
        return divider_idx + 1
    else:
        return divider_idx


def previous_divider_idx(dividers, minute_val):

    divider_idx = np.searchsorted(dividers, minute_val)

    if divider_idx == 0:
        raise ValueError("Cannot go earlier in calendar!")

    return divider_idx - 1


def compute_all_minutes(
    opens_in_ns, break_starts_in_ns, break_ends_in_ns, closes_in_ns,
):
    """
    Given arrays of opens and closes (in nanoseconds) and optionally
    break_starts and break ends, return an array of each minute between the
    opens and closes.

    NOTE: Add an extra minute to ending boundaries (break_start and close)
    so we include the last bar (arange doesn't include its stop).
    """
    pieces = []
    for open_time, break_start_time, break_end_time, close_time in zip(
        opens_in_ns, break_starts_in_ns, break_ends_in_ns, closes_in_ns
    ):
        if break_start_time != NP_NAT:
            pieces.append(
                np.arange(
                    open_time,
                    break_start_time + NANOSECONDS_PER_MINUTE,
                    NANOSECONDS_PER_MINUTE,
                )
            )
            pieces.append(
                np.arange(
                    break_end_time,
                    close_time + NANOSECONDS_PER_MINUTE,
                    NANOSECONDS_PER_MINUTE,
                )
            )
        else:
            pieces.append(
                np.arange(
                    open_time,
                    close_time + NANOSECONDS_PER_MINUTE,
                    NANOSECONDS_PER_MINUTE,
                )
            )
    out = np.concatenate(pieces).view("datetime64[ns]")
    return out



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    GoodFriday,
    EasterMonday,
)
from pytz import timezone, UTC

from .common_holidays import (
    new_years_day,
    epiphany,
    european_labour_day,
    corpus_christi,
    all_saints_day,
    christmas_eve,
    christmas,
    boxing_day,
    new_years_eve,
)
from .trading_calendar import HolidayCalendar, TradingCalendar


def not_2004(datetime_index):
    """
    Christmas Eve is a holiday every year except for whatever reason it was a
    trading day in 2004.
    """
    return datetime_index[datetime_index.year != 2004]


NewYearsDay = new_years_day()

Epiphany = epiphany(start_date='2011')

LabourDay = european_labour_day()

May3ConstitutionDay = Holiday(
    "Celabration of Declaration of the Constitution of 3 May",
    month=5,
    day=3,
)

CorpusChristi = corpus_christi()

ArmedForcesDay = Holiday(
    "Armed Forces Day",
    month=8,
    day=15,
)

AllSaintsDay = all_saints_day()

IndependenceDay = Holiday(
    "National Independence Day",
    month=11,
    day=11,
)

ChristmasEve = christmas_eve(observance=not_2004)
Christmas = christmas()
BoxingDay = boxing_day()

NewYearsEve = new_years_eve(start_date='2011')


class XWARExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Warsaw Stock Exchange (WSE).

    Open Time: 9:00 AM, Central European Time (CET)
    Close Time: 5:00 PM, Central European Time (CET)

    Regularly-Observed Holidays:
      - New Year's Day
      - Epiphany
      - Good Friday
      - Easter Monday
      - Labour Day
      - Constitution Day
      - Corpus Christi
      - Armed Forces Day
      - All Saints' Day
      - Independence Day
      - Christmas Eve
      - Christmas Day
      - Boxing Day
      - New Year's Eve

    Holidays No Longer Observed:
      - None

    Early Closes:
      - None
    """
    name = 'XWAR'

    tz = timezone('Europe/Warsaw')

    open_times = (
        (None, time(9, 1)),
    )
    close_times = (
        (None, time(17)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            Epiphany,
            GoodFriday,
            EasterMonday,
            LabourDay,
            May3ConstitutionDay,
            CorpusChristi,
            ArmedForcesDay,
            AllSaintsDay,
            IndependenceDay,
            ChristmasEve,
            Christmas,
            BoxingDay,
            NewYearsEve,
        ])

    @property
    def adhoc_holidays(self):
        return [
            pd.Timestamp('2005-04-08', tz=UTC),  # Pope's Funeral.
            pd.Timestamp('2007-12-31', tz=UTC),  # New Year's Eve (adhoc).
            pd.Timestamp('2008-05-02', tz=UTC),  # Exchange Holiday.
            pd.Timestamp('2009-01-02', tz=UTC),  # Exchange Holiday.
            pd.Timestamp('2013-04-16', tz=UTC),  # Exchange Holiday.
            pd.Timestamp('2018-01-02', tz=UTC),  # Exchange Holiday.
            pd.Timestamp('2018-11-12', tz=UTC),  # Independence Holiday.
        ]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta
from dateutil.easter import easter, EASTER_ORTHODOX
from itertools import chain
import pandas as pd
from pandas.tseries.holiday import (
    EasterMonday,
    GoodFriday,
    Holiday,
)
from pytz import timezone, UTC

from .common_holidays import (
    assumption_day,
    christmas,
    christmas_eve,
    epiphany,
    european_labour_day,
    new_years_day,
)
from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
)


def orthodox_easter(start_date='1980', end_date='2021'):
    """
    ASEX observes Orthodox Easter, and has many holidays
    that are relative to Orthodox Easter.  This function gives a
    DatetimeIndex of Orthodox Easter dates from start_date to end_date
    """
    return pd.to_datetime([
        easter(year, method=EASTER_ORTHODOX)
        for year in range(int(start_date), int(end_date))
    ]).tz_localize(UTC)


NewYearsDay = new_years_day()

Epiphany = epiphany()

OrthodoxAshMonday = orthodox_easter() - timedelta(48)

NationalHoliday1 = Holiday(
    'National Holiday 1',
    month=3,
    day=25,
)

OrthodoxGoodFriday = orthodox_easter() - timedelta(2)

OrthodoxEasterMonday = orthodox_easter() + timedelta(1)

LabourDay = european_labour_day()

OrthodoxWhitMonday = orthodox_easter() + timedelta(50)

AssumptionDay = assumption_day()

NationalHoliday2 = Holiday(
    'National Holiday 2',
    month=10,
    day=28,
)

ChristmasEve = christmas_eve(start_date='2009')

ChristmasDay = christmas()

SecondDayOfChristmas = Holiday(
    'Second Day of Christmas',
    month=12,
    day=26,
)


class ASEXExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Athens Stock Exchange (ASEX).

    Open Time: 10:00 AM, EEST
    Close Time: 5:00 PM, EEST (until 2008-09-29)
                5:20 PM, EEST (starting 2008-09-29)

    Regularly-Observed Holidays:
    - New Year's Day
    - Epiphany (Jan 6)
    - Orthodox Ash Monday (48 days before Orthodox Easter Sunday)
    - National Holiday (Mar 25)
    - Good Friday
    - Easter Monday
    - Orthodox Good Friday
    - Orthodox Easter Monday
    - Labour Day (May 1)
    - Orthodox Whit Monday (50 days after Orthodox Easter Sunday)
    - Assumption Day (Aug 15)
    - National Holiday (Oct 28)
    - Christmas Eve
    - Christmas Day
    - Second Day of Christmas (Dec 26)

    Early Closes:
    - None
    """
    name = 'ASEX'

    tz = timezone('Europe/Athens')

    open_times = (
        (None, time(10, 1)),
    )

    close_times = (
        (None, time(17, 00)),
        (pd.Timestamp('2008-09-29'), time(17, 20))
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            Epiphany,
            NationalHoliday1,
            GoodFriday,
            EasterMonday,
            LabourDay,
            AssumptionDay,
            NationalHoliday2,
            ChristmasEve,
            ChristmasDay,
            SecondDayOfChristmas
        ])

    @property
    def adhoc_holidays(self):
        debt_crisis = pd.date_range('2015-06-29', '2015-07-31', freq='B')

        debt_crisis_holidays = [
            pd.Timestamp(str(date), tz=UTC) for date in debt_crisis
        ]

        misc_adhoc_holidays = [
            # In 2002, market closed for unknown reason
            pd.Timestamp('2002-05-07', tz=UTC),
            # In 2004, Assumption Day fell on a sunday, observed on a Friday
            pd.Timestamp('2004-08-13', tz=UTC),
            # In 2008, worker strikes closed the market for 2 days in March
            pd.Timestamp('2008-03-04', tz=UTC),
            pd.Timestamp('2008-03-05', tz=UTC),
            # In 2013, May Day strikes closed the market
            pd.Timestamp('2013-05-07', tz=UTC),
            # In 2014, New Year's Eve was observed as a holiday
            pd.Timestamp('2014-12-31', tz=UTC),
            # In 2016, Labour Day fell on a Sunday, observed on Tuesday
            pd.Timestamp('2016-05-03', tz=UTC),
        ]

        return list(chain(
            debt_crisis_holidays,
            misc_adhoc_holidays,
            # TODO: Investigate making orthodox easter adhocs actual holidays
            OrthodoxGoodFriday,
            OrthodoxEasterMonday,
            OrthodoxWhitMonday,
            OrthodoxAshMonday,
        ))



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .trading_calendar import TradingCalendar
from .calendar_utils import (
    clear_calendars,
    deregister_calendar,
    get_calendar,
    get_calendar_names,
    register_calendar,
    register_calendar_alias,
    register_calendar_type,
    resolve_alias,
)

__all__ = [
    'clear_calendars',
    'deregister_calendar',
    'get_calendar',
    'get_calendar_names',
    'register_calendar',
    'register_calendar_alias',
    'register_calendar_type',
    'resolve_alias',
    'TradingCalendar',
]

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import (
    GoodFriday,
    Holiday,
)
from pytz import timezone

from .common_holidays import (
    ascension_day,
    christmas,
    new_years_day,
    new_years_eve,
)
from .lunisolar_holidays import (
    chinese_lunar_new_year_dates
)
from .trading_calendar import HolidayCalendar, TradingCalendar


class XIDXExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Indonesia Stock Exchange (XIDX).

    Open Time: 09:00, Western Indonesian Time (WIB)
    Close Time: 15:50, Western Indonesian Time (WIB)

    Regularly-Observed Holidays:
    - New Year's Day(Jan 1)
    - Good Friday (Friday before Easter)
    - Labor Day (May 1)
    - Ascension Day Of Jesus Christ (39 days after Easter, always Thursday)
    - Pancasila Day (Jun 1)
    - Independence Day (Aug 17)
    - Christmas Day (Dec 25)
    - New Year's Eve (called "Trading Holiday", Dec 31)
    - Chinese New Year (from Gregorian year 2002 onward)
    - Islamic New Year
    - Eid al-Fitr (Festival of Breaking the Fast)
    - Eid al-Adha (Festival of the Sacrifice)
    - Isra Mikraj of the Prophet Muhammad
    - Birth of the Prophet Muhammad
    - Vesak Day
    - Hindu Saka New Year (also called Nyepi, or Balinese Day of Silence)

    Election holidays are also observed, as well as additional "common
    leave" days around many holidays.
    """
    name = 'XIDX'

    tz = timezone('Asia/Jakarta')

    open_times = (
        (None, time(9, 1)),
    )

    close_times = (
        (None, time(15, 50)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            new_years_day(),
            GoodFriday,
            Holiday(
                "Labor Day",
                month=5,
                day=1,
                start_date='2014-05-01',
            ),
            # XXX: The Indonesia Stock Exchange was open on Ascension
            # Day 2003, but closed the next day. We are not sure why.
            # (The 2003-05-30 closure is listed in adhoc_holidays.)
            ascension_day(end_date='2003'),
            ascension_day(start_date='2004'),
            Holiday(
                "Pancasila Day",
                month=6,
                day=1,
                start_date='2017-06-01',
            ),
            Holiday(
                "Independence Day",
                month=8,
                day=17,
            ),
            # Christmas Eve is not an official Indonesian holiday, but
            # December 24th and 26th are frequently observed as common
            # leave. These dates are listed in adhoc_holidays.
            christmas(),
            new_years_eve(),
        ])

    chinese_new_year = chinese_lunar_new_year_dates[
        # The Indonesia Stock Exchange did not close for Chinese New
        # Year in 1998, 1999, or 2001. (It fell on a Saturday in 2000.)
        chinese_lunar_new_year_dates.year >= 2002
    ]

    common_leave = pd.to_datetime([
        # Indonesia sometimes observes additional "common leave" days
        # around the usual observed dates of holidays.

        # Common leave for New Year's Day.
        '2002-12-30',
        '2005-12-30',
        '2009-01-02',
        '2017-01-02',

        # Common leave for Chinese New Year.
        '2008-02-08',

        # Common leave for Ascension Day.
        '2003-05-30',
        '2006-05-26',
        '2007-05-18',

        # Common leave for Independence Day.
        '2003-08-18',
        '2006-08-18',
        '2008-08-18',

        # Common leave for Christmas.
        '2002-12-24',
        '2002-12-26',
        '2003-12-24',
        '2003-12-26',
        '2004-12-24',
        '2005-12-26',
        '2007-12-24',
        '2009-12-24',
        '2010-12-24',
        '2011-12-26',
        '2012-12-24',
        '2013-12-26',
        '2014-12-26',
        '2016-12-26',
        '2017-12-26',
        '2018-12-24',
        '2019-12-24',
        '2020-12-24',
    ])

    islamic_new_year = pd.to_datetime([
        # Islamic/Hijri/Hijriyah New Year.
        # Includes common leave.
        '2002-03-15',
        '2003-03-03',
        '2004-02-23',
        '2005-02-10',
        '2006-01-31',
        '2008-01-10',  # First Islamic New Year of 2008.
        '2008-01-11',
        '2008-12-29',  # Second Islamic New Year of 2008.
        '2009-12-18',
        '2010-12-07',
        '2012-11-15',
        '2012-11-16',
        '2013-11-05',
        '2015-10-14',
        '2017-09-21',
        '2018-09-11',
        '2020-08-20',
    ])

    eid_al_fitr = pd.to_datetime([
        # Eid al-Fitr (Festival of Breaking the Fast).
        # Includes common leave.
        '2002-12-05',
        '2002-12-06',
        '2002-12-09',
        '2002-12-10',

        '2003-11-24',
        '2003-11-25',
        '2003-11-26',
        '2003-11-27',
        '2003-11-28',

        '2004-11-15',
        '2004-11-16',
        '2004-11-17',
        '2004-11-18',
        '2004-11-19',

        '2005-11-02',
        '2005-11-03',
        '2005-11-04',
        '2005-11-07',
        '2005-11-08',

        '2006-10-23',
        '2006-10-24',
        '2006-10-25',
        '2006-10-26',
        '2006-10-27',

        '2007-10-12',
        '2007-10-15',
        '2007-10-16',

        '2008-09-30',
        '2008-10-01',
        '2008-10-02',
        '2008-10-03',

        '2009-09-18',
        '2009-09-21',
        '2009-09-22',
        '2009-09-23',

        '2010-09-08',
        '2010-09-09',
        '2010-09-10',
        '2010-09-13',
        '2010-09-14',

        '2011-08-29',
        '2011-08-30',
        '2011-08-31',
        '2011-09-01',
        '2011-09-02',

        '2012-08-20',
        '2012-08-21',
        '2012-08-22',

        '2013-08-05',
        '2013-08-06',
        '2013-08-07',
        '2013-08-08',
        '2013-08-09',

        '2014-07-28',
        '2014-07-29',
        '2014-07-30',
        '2014-07-31',
        '2014-08-01',

        '2015-07-16',
        '2015-07-17',
        '2015-07-20',
        '2015-07-21',

        '2016-07-04',
        '2016-07-05',
        '2016-07-06',
        '2016-07-07',
        '2016-07-08',

        '2017-06-23',
        '2017-06-26',
        '2017-06-27',
        '2017-06-28',
        '2017-06-29',
        '2017-06-30',

        '2018-06-11',
        '2018-06-12',
        '2018-06-13',
        '2018-06-14',
        '2018-06-15',
        '2018-06-18',
        '2018-06-19',

        '2019-06-03',
        '2019-06-04',
        '2019-06-05',
        '2019-06-06',
        '2019-06-07',

        '2020-05-22',
        '2020-05-25',
        '2020-05-26',
        '2020-05-27',
    ])

    eid_al_adha = pd.to_datetime([
        # Eid al-Adha (Festival of the Sacrifice).
        # Includes common leave.
        '2002-02-22',
        '2003-02-12',
        '2004-02-02',
        '2005-01-21',
        '2006-01-10',
        '2006-12-29',
        # NOTE: Eid al-Adha occured twice in 2006, on Tuesday 01-10 and
        # Sunday 12-31. The exchange was closed on Friday 2006-12-29 as
        # a make-up holiday.
        '2007-12-20',
        '2007-12-21',
        '2008-12-08',
        '2009-11-27',
        '2009-11-28',
        '2010-11-17',
        '2012-10-26',
        '2013-10-14',
        '2013-10-15',
        '2015-09-24',
        '2016-09-12',
        '2017-09-01',
        '2018-08-22',
        '2020-07-31',
    ])

    isra_mikraj = pd.to_datetime([
        # Isra and Mi'raj (Ascension of the Prophet Muhammad).
        # Called "Isra Mikraj" in Indonesia.
        #
        # Occurs on 27 Rajab on the Hijri calendar, but the mapping of
        # Hijri to Gregorian dates varies. For example, in 2018 many
        # countries observed this holiday on Friday 04-13; but by
        # Indonesian reckoning it fell on Saturday 04-14 that year.
        #
        # See https://www.idx.co.id/en-us/news/trading-holiday/
        #
        # Includes common leave.
        '2002-10-04',
        '2003-09-22',
        '2004-09-13',
        '2005-09-02',
        '2006-08-21',
        '2008-07-30',
        '2009-07-20',
        '2011-06-29',
        '2012-05-18',
        '2013-06-06',
        '2014-05-27',
        '2016-05-06',
        '2017-04-24',
        '2019-04-03',
    ])

    birth_of_prophet_muhammad = pd.to_datetime([
        # Birth of the Prophet Muhammad.
        # Includes common leave.
        '2003-05-15',
        '2004-05-03',
        '2005-04-22',
        '2006-04-10',
        '2008-03-20',
        '2009-03-09',
        '2010-02-26',
        '2011-02-15',
        '2013-01-24',
        '2014-01-14',
        '2015-12-24',
        '2016-12-12',
        '2017-12-01',
        '2018-11-20',
        '2020-10-29',
    ])

    vesak_day = pd.to_datetime([
        # Vesak Day (Buddha's Birthday).
        # Sometimes called "Hari Raya Waisak" in Indonesia.
        # Includes common leave.
        '2003-05-16',
        '2004-06-03',
        '2005-05-24',
        '2007-06-01',
        '2008-05-20',
        '2010-05-28',
        '2011-05-17',
        '2014-05-15',
        '2015-06-02',
        '2017-05-11',
        '2018-05-29',
        '2020-05-07',
    ])

    hindu_saka_new_year = pd.to_datetime([
        # Hindu Saka New Year (also called Nyepi, or Balinese Day of Silence).
        # Includes common leave.
        '2003-04-02',
        '2004-03-22',
        '2005-03-11',
        '2006-03-30',
        '2006-03-31',
        '2007-03-19',
        '2008-03-07',
        '2009-03-26',
        '2010-03-16',
        '2012-03-23',
        '2013-03-12',
        '2014-03-31',
        '2016-03-09',
        '2017-03-28',
        '2019-03-07',
        '2020-03-25',
    ])

    spontaneous_closures = pd.to_datetime([
        # Trading suspension due to global financial crisis.
        '2008-10-09',
        '2008-10-10',
    ])

    election_holidays = pd.to_datetime([
        # Local and gubernatorial election holidays.
        '2004-04-05',
        '2004-07-05',
        '2004-09-20',
        '2009-04-09',
        '2009-07-08',
        '2014-04-09',
        '2014-07-09',
        '2015-12-09',
        '2017-02-15',
        '2017-04-19',
        '2019-04-17',
    ])

    @property
    def adhoc_holidays(self):
        return list(chain(
            self.chinese_new_year,
            self.common_leave,
            self.islamic_new_year,
            self.eid_al_fitr,
            self.eid_al_adha,
            self.isra_mikraj,
            self.birth_of_prophet_muhammad,
            self.vesak_day,
            self.hindu_saka_new_year,
            self.spontaneous_closures,
            self.election_holidays,
        ))



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
import pandas as pd
from pandas.tseries.holiday import (
    Easter,
    EasterMonday,
    Holiday,
)
from pandas.tseries.offsets import Day
from pytz import UTC, timezone

from .common_holidays import (
    new_years_day,
    new_years_eve,
    european_labour_day,
    christmas,
    christmas_eve,
)
from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
)


NewYearsDay = new_years_day()

# Need custom start year so can't use pandas GoodFriday
GoodFriday = Holiday(
    "Good Friday",
    month=1,
    day=1,
    offset=[Easter(), Day(-2)],
    start_date='2013'
)

LabourDay = european_labour_day()

LiberationDay = Holiday(
    'Liberation Day',
    month=5,
    day=8,
)

SaintsCyrilAndMethodiusDay = Holiday(
    'Saints Cyril and Methodius Day',
    month=7,
    day=5,
)

JanHusDay = Holiday(
    'Jan Hus Day',
    month=7,
    day=6,
)

CzechStatehoodDay = Holiday(
    'Czech Statehood Day',
    month=9,
    day=28,
)

IndependenceDay = Holiday(
    'Independence Day',
    month=10,
    day=28,
)

StruggleForFreedomAndDemocracyDay = Holiday(
    'Struggle for Freedom and Democracy Day',
    month=11,
    day=17,
)

ChristmasEve = christmas_eve()

ChristmasDay = christmas()

SecondDayOfChristmas = Holiday(
    'Second Day of Christmas',
    month=12,
    day=26,
)

ExchangeHoliday = new_years_eve()


class XPRAExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Prague Stock Exchange (XPRA).

    Open Time: 9:00 AM, CET
    Close Time: 4:20 PM, CET

    Regularly-Observed Holidays:
    - New Years Day
    - Good Friday
    - Easter Monday
    - Labour Day (May 1)
    - Liberation Day (May 8)
    - Saints Cyril and Methodius Day (Jul 5)
    - Jan Hus Day (Jul 6)
    - Czech Statehood Day (Sep 28)
    - Independence Day (Oct 28)
    - Struggle for Freedom and Democracy Day (Nov 17)
    - Christmas Eve
    - Christmas Day
    - Second Day of Christmas (Dec 26)
    - Exchange Holiday (Dec 31)

    Early Closes:
    - None
    """
    name = 'XPRA'

    tz = timezone('Europe/Prague')

    open_times = (
        (None, time(9, 1)),
    )

    close_times = (
        (None, time(16, 20)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            GoodFriday,
            EasterMonday,
            LabourDay,
            LiberationDay,
            SaintsCyrilAndMethodiusDay,
            JanHusDay,
            CzechStatehoodDay,
            IndependenceDay,
            StruggleForFreedomAndDemocracyDay,
            ChristmasEve,
            ChristmasDay,
            SecondDayOfChristmas,
            ExchangeHoliday,
        ])

    @property
    def adhoc_holidays(self):
        return [
            # Extreme Flooding
            pd.Timestamp('2002-08-14', tz=UTC),
            # Restoration of the Czech Independence Day
            pd.Timestamp('2004-01-02', tz=UTC),
            pd.Timestamp('2005-01-03', tz=UTC),
        ]



from abc import abstractproperty
from pytz import UTC
import numpy as np
import pandas as pd
from .trading_calendar import TradingCalendar


class PrecomputedTradingCalendar(TradingCalendar):
    """
    Used to model an exchange calendar whose holidays are precomputed and
    hardcoded.
    """
    def __init__(self, start=None, end=None):
        earliest_precomputed_year = np.min(self.precomputed_holidays).year
        latest_precomputed_year = np.max(self.precomputed_holidays).year

        if start is None:
            start = pd.Timestamp(
                '{}-01-01'.format(earliest_precomputed_year), tz=UTC
            )

        if end is None:
            end = pd.Timestamp(
                '{}-12-31'.format(latest_precomputed_year), tz=UTC
            )

        super(PrecomputedTradingCalendar, self).__init__(start=start, end=end)

        if earliest_precomputed_year > self.first_trading_session.year:
            raise ValueError(
                'The {} holidays are only recorded back to {},'
                ' cannot instantiate the {} calendar back to {}.'.format(
                    self.name,
                    earliest_precomputed_year,
                    self.name,
                    self.first_trading_session.year
                ),
            )

        if latest_precomputed_year < self.last_trading_session.year:
            raise ValueError(
                'The {} holidays are only recorded to {},'
                ' cannot instantiate the {} calendar for {}.'.format(
                    self.name,
                    latest_precomputed_year,
                    self.name,
                    self.last_trading_session.year
                ),
            )

    @abstractproperty
    def precomputed_holidays(self):
        raise NotImplementedError()

    @property
    def adhoc_holidays(self):
        return self.precomputed_holidays



#!/usr/bin/env python
#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import (
    find_packages,
    setup,
)

import versioneer

DISTNAME = "trading_calendars"
DESCRIPTION = """trading_calendars is a Python library with \
securities exchange calendars used by Quantopian's Zipline."""

AUTHOR = "Quantopian Inc"
AUTHOR_EMAIL = "opensource@quantopian.com"
URL = "https://github.com/quantopian/trading_calendars"
LICENSE = "Apache License, Version 2.0"

classifiers = [
    "Development Status :: 4 - Beta",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "License :: OSI Approved :: Apache Software License",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Operating System :: OS Independent"
]

reqs = [
    "numpy",
    "pandas",
    "pytz",
    "toolz",
]

with open('README.md') as f:
    LONG_DESCRIPTION = f.read()

if __name__ == '__main__':
    setup(
        name=DISTNAME,
        entry_points={
            'console_scripts': [
                'tcal = trading_calendars.tcal:main',
            ],
        },
        cmdclass=versioneer.get_cmdclass(),
        version=versioneer.get_version(),
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        description=DESCRIPTION,
        license=LICENSE,
        url=URL,
        classifiers=classifiers,
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        packages=find_packages(
            include=['trading_calendars', 'trading_calendars.*']
        ),
        install_requires=reqs,
        extras_require={
            "test": [
                "flake8",
                "nose",
                "nose-ignore-docstring",
                "nose-timer",
                "parameterized",
            ],
        },
    )



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta
from pandas.tseries.holiday import (
    Easter,
    EasterMonday,
    Holiday,
)
from pandas.tseries.offsets import Day
from pytz import timezone

from .common_holidays import (
    new_years_day,
    new_years_eve,
    european_labour_day,
    christmas,
    christmas_eve,
    all_saints_day,
    whit_monday,
)
from .trading_calendar import (
    TradingCalendar,
    HolidayCalendar,
    TUESDAY,
    THURSDAY,
)


def four_day_weekend(dt):
    """
    For almost all holidays in the XBUD calendar, if the holiday
    falls on a Tuesday the previous Monday also becomes a holiday,
    and if the holiday falls on a Thursday the following Friday also
    becomes a holiday.
    """
    mon = dt[dt.weekday == TUESDAY] - timedelta(1)  # mv Tues back one day
    fri = dt[dt.weekday == THURSDAY] + timedelta(1)  # mv Thurs ahead one day
    return dt.append([mon, fri])


NewYearsDay = new_years_day(observance=four_day_weekend)

NationalHoliday1 = Holiday(
    'National Holiday 1',
    month=3,
    day=15,
    observance=four_day_weekend
)

# Need custom start year so can't use pandas GoodFriday
GoodFriday = Holiday(
    "Good Friday",
    month=1,
    day=1,
    offset=[Easter(), Day(-2)],
    start_date='2012'
)

LabourDay = european_labour_day(observance=four_day_weekend)

WhitMonday = whit_monday()

StStephensDay = Holiday(
    "St. Stephen's Day",
    month=8,
    day=20,
    observance=four_day_weekend,
)

NationalHoliday2 = Holiday(
    'National Holiday 2',
    month=10,
    day=23,
    observance=four_day_weekend,
)

AllSaintsDay = all_saints_day(observance=four_day_weekend)

# Christmas Eve does not follow the four day weekend rule
ChristmasEve = christmas_eve()

ChristmasDay = christmas()

# XBUD always has a holiday for the second day of Christmas (26th),
# but starting in 2013 if the 26th falls on a Thursday then the
# 27th (Friday) is also taken off
SecondDayOfChristmas = Holiday(
    'Second Day of Christmas (w/ no added Friday off)',
    month=12,
    day=26,
    end_date='2013',
)

SecondDayOfChristmasAddFriday = Holiday(
    'Second Day of Christmas (w/ added Friday off)',
    month=12,
    day=26,
    start_date='2013',
    observance=four_day_weekend,
)

# Starting in 2011, New Year's Eve is observed as a holiday every year.
# In some cases pre-2011, the 31st becomes a holiday due to the four day
# weekend rule (when Jan 1 falls on a Tuesday).
# Also, when NYE starts being observed as a holiday it does NOT follow
# the four day weekend rule (no 30ths are holidays)
NewYearsEve = new_years_eve(start_date='2011')


class XBUDExchangeCalendar(TradingCalendar):
    """
    Exchange calendar for the Budapest Stock Exchange (XBUD).

    Open Time: 9:00 AM, CET
    Close Time: 5:00 PM, CET

    Regularly-Observed Holidays:
    - New Year's Day
    - National Holiday (Mar 15)
    - Good Friday
    - Easter Monday
    - Labour Day (May 1)
    - Whit Monday (50 days after Easter Sunday)
    - St. Stephen's Day (Aug 20)
    - National Holiday (Oct 23)
    - All Saint's Day (Nov 1)
    - Christmas Eve
    - Christmas Day
    - Second Day of Christmas (Dec 26)
    - New Year's Eve

    Early Closes:
    - None
    """
    name = 'XBUD'

    tz = timezone('Europe/Budapest')

    open_times = (
        (None, time(9, 1)),
    )

    close_times = (
        (None, time(17, 00)),
    )

    @property
    def regular_holidays(self):
        return HolidayCalendar([
            NewYearsDay,
            NationalHoliday1,
            GoodFriday,
            EasterMonday,
            LabourDay,
            WhitMonday,
            StStephensDay,
            NationalHoliday2,
            AllSaintsDay,
            ChristmasEve,
            ChristmasDay,
            SecondDayOfChristmas,
            SecondDayOfChristmasAddFriday,
            NewYearsEve,
        ])
